# Author:  Lisandro Dalcin
# Contact: dalcinl@gmail.com
"""Support for MPI ABI."""

import importlib.machinery
import importlib.util
import os
import sys
import warnings

MPIABI = None  # type: str | None
LIBMPI = None  # type: str | None
LIBMPI_PATH = []  # type: list[str]
LIBMPI_MODE = None  # type: int | None


def _verbose_info(message, verbosity=1):
    if sys.flags.verbose >= verbosity:
        print(f"# [{__name__}] {message}", file=sys.stderr)


def _site_prefixes():
    prefixes = []
    site = sys.modules.get("site")
    if site is not None:  # pragma: no cover
        if sys.exec_prefix != sys.base_exec_prefix:
            venv_base = sys.exec_prefix
            prefixes.append(venv_base)
        if site.ENABLE_USER_SITE:
            user_base = os.path.abspath(site.USER_BASE)
            prefixes.append(user_base)
        if sys.base_exec_prefix in site.PREFIXES:
            system_base = sys.base_exec_prefix
            prefixes.append(system_base)
    return prefixes


def _dlopen_rpath():
    rpath = []

    def add_rpath(*directory):
        path = os.path.join(*directory)
        if path not in rpath:
            rpath.append(path)

    def add_rpath_prefix(prefix):
        if os.name == "posix":
            if prefix != "/usr":  # pragma: no branch
                add_rpath(prefix, "lib")
        else:
            add_rpath(prefix, "DLLs")
            add_rpath(prefix, "Library", "bin")

    for prefix in _site_prefixes():
        add_rpath_prefix(prefix)

    add_rpath("")

    if os.name == "nt":
        impi_root = os.environ.get("I_MPI_ROOT")
        impi_library_kind = os.environ.get("I_MPI_LIBRARY_KIND") or "release"
        msmpi_bin = os.environ.get("MSMPI_BIN")
        if not msmpi_bin:
            msmpi_root = os.environ.get("MSMPI_ROOT")
            if msmpi_root:
                msmpi_bin = os.path.join(msmpi_root, "bin")
        if impi_root:
            add_rpath(impi_root, "bin", "mpi", impi_library_kind)
            add_rpath(impi_root, "bin", impi_library_kind)
            add_rpath(impi_root, "bin")
        if msmpi_bin:
            add_rpath(msmpi_bin)

    if sys.platform == "darwin":
        add_rpath("/usr/local/lib")
        add_rpath("/opt/homebrew/lib")
        add_rpath("/opt/local/lib")

    return rpath


def _dlopen_mode():  # pragma: no cover
    if sys.platform == "linux":
        return os.RTLD_LAZY | os.RTLD_LOCAL
    if sys.platform == "darwin":
        return os.RTLD_LAZY | os.RTLD_GLOBAL
    if os.name == "posix":
        return os.RTLD_LAZY | os.RTLD_LOCAL
    return None


def _dlopen_libmpi(libmpi=None):  # pragma: no cover
    # pylint: disable=too-many-statements
    # pylint: disable=import-outside-toplevel
    import ctypes as ct

    def dlopen(name, mode=None):
        if mode is None:
            mode = ct.DEFAULT_MODE
        _verbose_info(f"trying to dlopen {name!r}")
        lib = ct.CDLL(name, mode)
        _ = lib.MPI_Get_version
        _verbose_info(f"MPI library from {name!r}")
        if name is not None and sys.platform == "linux":
            if hasattr(lib, "I_MPI_Check_image_status"):
                if os.path.basename(name) != name:
                    dlopen_impi_libfabric(os.path.dirname(name), mode)
        if name is not None and os.name == "nt":
            if os.path.basename(name).lower() == "impi.dll":
                if os.path.basename(name) != name:
                    dlopen_impi_libfabric(os.path.dirname(name), mode)
        return lib

    def search_impi_libfabric(rootdir):
        if sys.platform == "linux":
            libdir = "lib"
            suffix = ".so.1"
        else:
            libdir = "bin"
            suffix = ".dll"
        for subdir in (
            ("opt", "mpi", "libfabric", libdir),
            ("libfabric", libdir),
            ("libfabric",),
            (libdir, "libfabric"),
            (libdir,),
        ):
            ofi_libdir = os.path.join(rootdir, *subdir)
            ofi_filename = os.path.join(ofi_libdir, f"libfabric{suffix}")
            if os.path.isfile(ofi_filename):
                return ofi_filename
        return None

    def dlopen_impi_libfabric(libdir, mode):
        ofi_library_internal = os.environ.get(
            "I_MPI_OFI_LIBRARY_INTERNAL", ""
        ).lower() not in ("0", "no", "off", "false", "disable")
        ofi_required = os.environ.get("I_MPI_FABRICS") != "shm"
        if not (ofi_library_internal and ofi_required):
            return None
        rootdir = os.path.dirname(libdir)
        if os.path.basename(rootdir).lower() in ("release", "debug"):
            rootdir = os.path.dirname(rootdir)
        ofi_filename = search_impi_libfabric(rootdir)
        if ofi_filename is None:
            return None
        if "FI_PROVIDER_PATH" not in os.environ:
            ofi_libdir = os.path.dirname(ofi_filename)
            ofi_prov = os.path.join(ofi_libdir, "prov")
            ofi_path = ofi_prov if os.path.isdir(ofi_prov) else ofi_libdir
            os.environ["FI_PROVIDER_PATH"] = ofi_path
        _verbose_info(f"trying to dlopen {ofi_filename!r}")
        lib = ct.CDLL(ofi_filename, mode)
        _ = lib.fi_getinfo
        _verbose_info(f"OFI library from {ofi_filename!r}")
        return lib

    def libname(name, version=None):
        suffix = f".{version}" if version is not None else ""
        if sys.platform == "darwin":
            template = "lib{}{}.dylib"
        elif os.name == "posix":
            template = "lib{}.so{}"
        else:
            template = "{}.dll"
        return template.format(name, suffix)

    def libmpi_names():
        if os.name == "posix":
            yield libname("mpi")
            yield libname("mpi", 12)  # mpich
            yield libname("mpi", 40)  # openmpi
        else:
            yield libname("impi")
            yield libname("msmpi")

    def libmpi_paths(path):
        rpath = "@rpath" if sys.platform == "darwin" else ""
        for entry in path:
            entry = entry or rpath
            entry = os.path.expandvars(entry)
            entry = os.path.expanduser(entry)
            if entry == rpath or os.path.isdir(entry):
                for name in libmpi_names():
                    yield os.path.join(entry, name)
            else:
                yield entry

    if os.name == "posix":
        try:
            return dlopen(None)
        except (OSError, AttributeError):
            pass
    if libmpi is not None:
        path = libmpi.split(os.pathsep)
    else:
        path = LIBMPI_PATH or _dlopen_rpath() or [""]
    if LIBMPI_MODE is not None:
        mode = LIBMPI_MODE
    else:
        mode = _dlopen_mode()
    errors = ["cannot load MPI library"]
    for filename in libmpi_paths(path):
        try:
            return dlopen(filename, mode)
        except OSError as exc:
            errors.append(str(exc))
        except AttributeError as exc:
            errors.append(str(exc))
    raise RuntimeError("\n".join(errors))


def _get_mpiabi_from_libmpi(libmpi=None):
    # pylint: disable=import-outside-toplevel
    import ctypes as ct

    lib = _dlopen_libmpi(libmpi)
    abi_get_version = getattr(lib, "MPI_Abi_get_version", None)
    if abi_get_version:  # pragma: no cover
        abi_get_version.restype = ct.c_int
        abi_get_version.argtypes = [ct.POINTER(ct.c_int)] * 2
        abi_major, abi_minor = ct.c_int(0), ct.c_int(0)
        ierr = abi_get_version(ct.byref(abi_major), ct.byref(abi_minor))
        if ierr:  # pragma: no cover
            message = f"MPI_Abi_get_version [ierr={ierr}]"
            raise RuntimeError(message)
        if abi_major.value > 0:
            return "mpiabi"
    if os.name == "posix":  # pragma: no branch
        openmpi = hasattr(lib, "ompi_mpi_comm_self")
        mpiabi = "openmpi" if openmpi else "mpich"
    else:  # pragma: no cover
        msmpi = hasattr(lib, "MSMPI_Get_version")
        mpiabi = "msmpi" if msmpi else "impi"
    return mpiabi


def _get_mpiabi_from_string(string):
    table = {ord(c): "" for c in " -_"}
    mpiabi = string.translate(table).lower()
    if os.name == "posix":  # pragma: no branch
        if mpiabi == "impi":
            mpiabi = "mpich"
    else:  # pragma: no cover
        if mpiabi == "mpich":
            mpiabi = "impi"
    return mpiabi


def _get_mpiabi():
    mpiabi = getattr(_get_mpiabi, "mpiabi", None)
    if mpiabi is None:
        mpiabi = MPIABI or os.environ.get("MPI4PY_MPIABI")
        libmpi = LIBMPI or os.environ.get("MPI4PY_LIBMPI")
        if mpiabi is not None:
            mpiabi = _get_mpiabi_from_string(mpiabi)
        else:
            mpiabi = _get_mpiabi_from_libmpi(libmpi)
        _get_mpiabi.mpiabi = mpiabi  # pyright: ignore
    return mpiabi


_registry = {}  # type: dict[str, list[str]]


def _register(module, mpiabi):
    mpiabi = _get_mpiabi_from_string(mpiabi)
    registered = _registry.setdefault(module, [])
    if mpiabi not in registered:
        registered.append(mpiabi)


def _get_mpiabi_suffix(module):
    if module not in _registry:
        return None
    mpiabi = _get_mpiabi()
    if mpiabi not in _registry[module]:
        return None
    return f".{mpiabi}" if mpiabi else ""


class _Finder:
    """MPI ABI-aware extension module finder."""

    # pylint: disable=too-few-public-methods
    @classmethod
    def find_spec(cls, fullname, path, target=None):  # noqa: ARG003
        """Find MPI ABI extension module spec."""
        # pylint: disable=unused-argument
        mpiabi_suffix = _get_mpiabi_suffix(fullname)
        if mpiabi_suffix is None:
            return None
        _verbose_info(f"MPI ABI extension module: {fullname!r}")
        _verbose_info(f"MPI ABI extension suffix: {mpiabi_suffix!r}")
        ext_name = fullname.rpartition(".")[2]
        extension_suffixes = importlib.machinery.EXTENSION_SUFFIXES
        spec_from_file_location = importlib.util.spec_from_file_location
        for entry in path:
            for ext_suffix in extension_suffixes:
                filename = f"{ext_name}{mpiabi_suffix}{ext_suffix}"
                location = os.path.join(entry, filename)
                if os.path.isfile(location):
                    return spec_from_file_location(fullname, location)
        warnings.warn(
            f"unsupported MPI ABI {mpiabi_suffix[1:]!r}",
            category=RuntimeWarning,
            stacklevel=2,
        )
        return None


def _install_finder():
    if _Finder not in sys.meta_path:
        sys.meta_path.append(_Finder)



#!/usr/bin/env python

if False:
    import mpi4py

    name = "name"  # lib{name}.so
    path = []
    mpi4py.profile(name, path=path)

import threading
from array import array

from mpi4py import MPI

send_msg = array("i", [7] * 1000)
send_msg *= 1000
recv_msg = array("i", [0] * 1000)
recv_msg *= 1000


def self_send(comm, rank):
    comm.Send([send_msg, MPI.INT], dest=rank, tag=0)


def self_recv(comm, rank):
    comm.Recv([recv_msg, MPI.INT], source=rank, tag=0)


comm = MPI.COMM_WORLD
rank = comm.Get_rank()
send_thread = threading.Thread(target=self_send, args=(comm, rank))
recv_thread = threading.Thread(target=self_recv, args=(comm, rank))

send_thread.start()
recv_thread.start()
recv_thread.join()
send_thread.join()



import math
import sys

from mpi4py.futures import MPIPoolExecutor, get_comm_workers, wait


def compute_pi(n):
    comm = get_comm_workers()
    comm.barrier()

    n = comm.bcast(n, root=0)
    h = 1.0 / n
    s = 0.0
    for i in range(comm.rank + 1, n + 1, comm.size):
        x = h * (i - 0.5)
        s += 4.0 / (1.0 + x**2)
    pi = comm.allreduce(s * h)
    return pi


def main():
    try:
        n = int(sys.argv[1])
    except IndexError:
        n = 256
    try:
        P = int(sys.argv[2])
    except IndexError:
        P = 5
    with MPIPoolExecutor(P) as executor:
        P = executor.num_workers
        fs = [executor.submit(compute_pi, n) for _ in range(P)]
        wait(fs)
        pi = fs[0].result()
        print(
            f"pi: {pi:.15f}, error: {abs(pi - math.pi):.3e}",
            f"({n:d} intervals, {P:d} workers)",
        )


if __name__ == "__main__":
    main()



import sys

from mpi4py.futures import MPIPoolExecutor


def main():
    def getarg(opt, default=None):
        try:
            return sys.argv[sys.argv.index("--" + opt) + 1]
        except ValueError:
            return default

    if "--host" in sys.argv or "--port" in sys.argv:
        service = (getarg("host"), getarg("port"))
    else:
        service = getarg("service")
    if "--info" in sys.argv:
        info = getarg("info").split(",")
        info = dict(entry.split("=") for entry in info if entry)
    else:
        info = None

    with MPIPoolExecutor(service=service, mpi_info=info) as executor:
        fut1 = executor.submit(abs, +42)
        fut2 = executor.submit(abs, -42)
    assert fut1.result(0) == 42
    assert fut2.result(0) == 42


if __name__ == "__main__":
    main()



# Author:  Lisandro Dalcin
# Contact: dalcinl@gmail.com
"""The **MPI for Python** package.

The *Message Passing Interface* (MPI) is a standardized and portable
message-passing system designed to function on a wide variety of
parallel computers. The MPI standard defines the syntax and semantics
of library routines and allows users to write portable programs in the
main scientific programming languages (Fortran, C, or C++). Since its
release, the MPI specification has become the leading standard for
message-passing libraries for parallel computers.

*MPI for Python* provides MPI bindings for the Python programming
language, allowing any Python program to exploit multiple processors.
This package build on the MPI specification and provides an object
oriented interface which closely follows MPI-2 C++ bindings.

"""

__version__ = "4.1.1"
__author__ = "Lisandro Dalcin"
__credits__ = "MPI Forum, MPICH Team, Open MPI Team"


__all__ = ["MPI"]  # noqa: F822


class Rc:
    """Runtime configuration options.

    Attributes
    ----------
    initialize : bool
        Automatic MPI initialization at import (default: True).
    threads : bool
        Request initialization with thread support (default: True).
    thread_level : {"multiple", "serialized", "funneled", "single"}
        Level of thread support to request (default: "multiple").
    finalize : None or bool
        Automatic MPI finalization at exit (default: None).
    fast_reduce : bool
        Use tree-based reductions for objects (default: True).
    recv_mprobe : bool
        Use matched probes to receive objects (default: True).
    irecv_bufsz : int
        Default buffer size in bytes for ``irecv()`` (default = 32768).
    errors : {"exception", "default", "abort", "fatal"}
        Error handling policy (default: "exception").

    """

    initialize = True
    threads = True
    thread_level = "multiple"
    finalize = None
    fast_reduce = True
    recv_mprobe = True
    irecv_bufsz = 32768
    errors = "exception"

    def __init__(self, **kwargs):
        """Initialize options."""
        self(**kwargs)

    def __setattr__(self, name, value):
        """Set option."""
        if not hasattr(self, name):
            raise TypeError(f"object has no attribute {name!r}")
        super().__setattr__(name, value)

    def __call__(self, **kwargs):
        """Update options."""
        for key in kwargs:
            if not hasattr(self, key):
                raise TypeError(f"unexpected argument {key!r}")
        for key, value in kwargs.items():
            setattr(self, key, value)

    def __repr__(self):
        """Return repr(self)."""
        return f"<{__spec__.name}.rc>"


rc = Rc()
__import__("sys").modules[__spec__.name + ".rc"] = rc


def get_include():
    """Return the directory in the package that contains header files.

    Extension modules that need to compile against mpi4py should use
    this function to locate the appropriate include directory. Using
    Python distutils (or perhaps NumPy distutils)::

      import mpi4py
      Extension('extension_name', ...
                include_dirs=[..., mpi4py.get_include()])

    """
    prefix = __import__("pathlib").Path(__spec__.origin).parent
    return str(prefix / "include")


def get_config():
    """Return a dictionary with information about MPI.

    .. versionchanged:: 4.0.0
       By default, this function returns an empty dictionary. However,
       downstream packagers and distributors may alter such behavior.
       To that end, MPI information must be provided under an ``mpi``
       section within a UTF-8 encoded INI-style configuration file
       :file:`mpi.cfg` located at the top-level package directory.
       The configuration file is read and parsed using the
       `configparser` module.

    """
    prefix = __import__("pathlib").Path(__spec__.origin).parent
    parser = __import__("configparser").ConfigParser()
    parser.add_section("mpi")
    parser.read(prefix / "mpi.cfg", encoding="utf-8")
    return dict(parser.items("mpi"))


def profile(name, *, path=None):
    """Support for the MPI profiling interface.

    Parameters
    ----------
    name : str
       Name of the profiler library to load.
    path : `sequence` of str, optional
       Additional paths to search for the profiler.

    """
    # pylint: disable=import-outside-toplevel
    import os
    import pathlib
    import sys
    import warnings

    try:
        from _ctypes import dlopen
        from os import RTLD_GLOBAL, RTLD_NOW
    except ImportError as exc:  # pragma: no cover
        warnings.warn(exc.args[0], stacklevel=2)
        return

    def find_library(name, path):
        pattern = [("", "")]
        if sys.platform == "darwin":  # pragma: no cover
            pattern.append(("lib", ".dylib"))
        elif os.name == "posix":  # pragma: no cover
            pattern.append(("lib", ".so"))
        for pth in map(pathlib.Path, path):
            for lib, dso in pattern:
                filename = pth / f"{lib}{name}{dso}"
                if filename.is_file():
                    return str(filename.resolve())
        return None

    if path is None:
        path = [""]
    elif isinstance(path, os.PathLike):
        path = [path]
    elif isinstance(path, str):
        path = path.split(os.pathsep)
    elif isinstance(path, bytes):
        path = path.split(os.fsencode(os.pathsep))

    name = os.fsdecode(name)
    path = list(map(os.fsdecode, path))
    filename = find_library(name, path)
    if filename is None:
        raise ValueError(f"profiler {name!r} not found")

    try:
        handle = dlopen(filename, RTLD_NOW | RTLD_GLOBAL)
    except OSError as exc:
        warnings.warn(exc.args[0], stacklevel=2)
    else:
        registry = vars(profile).setdefault("registry", [])
        registry.append((name, (handle, filename)))



# -----------------------------------------------------------------------------

cdef extern from "Python.h":
    enum: PY_SSIZE_T_MAX

    void *PyMem_Malloc(size_t) noexcept
    void *PyMem_Calloc(size_t, size_t) noexcept
    void *PyMem_Realloc(void*, size_t) noexcept
    void  PyMem_Free(void*) noexcept

    void *PyMem_RawMalloc(size_t) noexcept nogil
    void *PyMem_RawCalloc(size_t, size_t) noexcept nogil
    void *PyMem_RawRealloc(void*, size_t) noexcept nogil
    void  PyMem_RawFree(void*) noexcept nogil


@cython.final
@cython.internal
cdef class _PyMem:

    cdef void *buf
    cdef Py_ssize_t len
    cdef void (*free)(void*) noexcept

    def __cinit__(self):
        self.buf = NULL
        self.len = 0
        self.free = NULL

    def __dealloc__(self):
        if self.free:
            self.free(self.buf)

    def __getbuffer__(self, Py_buffer *view, int flags):
        PyBuffer_FillInfo(view, self, self.buf, self.len, 0, flags)


cdef inline _PyMem allocate(Py_ssize_t m, size_t b, void *buf):
    if m > PY_SSIZE_T_MAX // <Py_ssize_t>b:
        raise MemoryError("memory allocation size too large")  # ~> uncovered
    if m < 0:
        raise RuntimeError("memory allocation with negative size")  # ~> uncovered
    cdef _PyMem ob = <_PyMem>New(_PyMem)
    ob.len  = m * <Py_ssize_t>b
    ob.free = PyMem_Free
    ob.buf  = PyMem_Malloc(<size_t>m * b)
    if ob.buf == NULL: raise MemoryError
    if buf != NULL: (<void**>buf)[0] = ob.buf
    return ob


cdef inline _PyMem rawalloc(Py_ssize_t m, size_t b, bint clear, void *buf):
    if m > PY_SSIZE_T_MAX // <Py_ssize_t>b:
        raise MemoryError("memory allocation size too large")  # ~> uncovered
    if m < 0:
        raise RuntimeError("memory allocation with negative size")  # ~> uncovered
    cdef _PyMem ob = <_PyMem>New(_PyMem)
    ob.len = m * <Py_ssize_t>b
    ob.free = PyMem_RawFree
    if clear:
        ob.buf = PyMem_RawCalloc(<size_t>m, b)
    else:
        ob.buf = PyMem_RawMalloc(<size_t>m * b)
    if ob.buf == NULL: raise MemoryError
    if buf != NULL: (<void**>buf)[0] = ob.buf
    return ob

# -----------------------------------------------------------------------------



from urllib.request import urlopen

from mpi4py.futures import MPIPoolExecutor

URLS = [
    "http://www.google.com/",
    "http://www.apple.com/",
    "http://www.ibm.com/",
    "http://www.slashdot.org/",
    "http://www.python.org/",
    "http://www.bing.com/",
    "http://www.facebook.com/",
    "http://www.github.com/",
    "http://www.youtube.com/",
    "http://www.blogger.com/",
]


def load_url(url):
    return url, urlopen(url).read()


def test_crawl():
    with MPIPoolExecutor(10) as executor:
        for url, content in executor.map(
            load_url, URLS, timeout=10, unordered=True
        ):
            kib = len(content) / (1 << 10)
            print(f"{url:25s}: {kib:6.2f} KiB")


if __name__ == "__main__":
    test_crawl()



import contextlib
import importlib
import os
import sys


def github():
    return os.environ.get("GITHUB_ACTIONS") == "true"


def azure():
    return os.environ.get("TF_BUILD") == "True"


def import_MPI():
    return importlib.import_module("mpi4py.MPI")


def has_datatype(datatype):
    # https://github.com/pmodels/mpich/issues/7341
    MPI = import_MPI()
    if datatype == MPI.DATATYPE_NULL:
        return False
    try:
        size = datatype.Get_size()
    except MPI.Exception as exc:
        if exc.Get_error_class() != MPI.ERR_TYPE:
            raise
        return False
    if size in (0, MPI.UNDEFINED):
        return False
    return True


def appnum():
    MPI = import_MPI()
    if MPI.APPNUM != MPI.KEYVAL_INVALID:
        return MPI.COMM_WORLD.Get_attr(MPI.APPNUM)
    return None


def has_mpi_appnum():
    return appnum() is not None


def has_mpi_port():
    MPI = import_MPI()
    try:
        port = MPI.Open_port()
    except MPI.Exception:
        return False
    try:
        MPI.Close_port(port)
    except MPI.Exception:
        return False
    return True


def disable_mpi_spawn():
    MPI = import_MPI()
    skip_spawn = os.environ.get("MPI4PY_TEST_SPAWN") in (
        "0",
        "n",
        "no",
        "off",
        "false",
    )
    if skip_spawn:
        return True
    macos = sys.platform == "darwin"
    windows = sys.platform == "win32"
    name, version = MPI.get_vendor()
    if name == "Open MPI":
        if version < (3, 0, 0):
            return True
        if version == (4, 0, 0):
            return True
        if version == (4, 0, 1) and macos:
            return True
        if version == (4, 0, 2) and macos:
            return True
        if (4, 1, 0) <= version < (4, 2, 0):
            if azure() or github():
                return True
    if name == "MPICH":
        if (3, 4, 0) <= version < (4, 0, 0):
            if macos:
                return True
        if version < (4, 1):
            if not has_mpi_appnum():
                return True
        if version < (4, 3):
            try:
                port = MPI.Open_port()
                MPI.Close_port(port)
            except MPI.Exception:
                return True
    if name == "Intel MPI":
        mpi4py = __import__("mpi4py")
        if mpi4py.rc.recv_mprobe:
            return True
        if MPI.COMM_WORLD.Get_size() > 1 and windows:
            return True
    if name == "Microsoft MPI":
        if version < (8, 1, 0):
            return True
        if has_mpi_appnum() is not None:
            return True
        if os.environ.get("PMI_APPNUM") is None:
            return True
    if name == "MVAPICH":
        if version < (3, 0, 0):
            return True
        if not has_mpi_appnum():
            return True
    if name == "MPICH2":
        if not has_mpi_appnum():
            return True
    if MPI.Get_version() < (2, 0):
        return True
    if any(map(sys.modules.get, ("cupy", "numba"))):
        return True
    #
    return False


@contextlib.contextmanager
def capture_stderr():
    stderr = sys.stderr
    stream = __import__("io").StringIO()
    sys.stderr = stream
    try:
        yield stream
    finally:
        sys.stderr = stderr



"""
Compare the speed of downloading URLs sequentially vs. using futures.
"""

import contextlib
import functools
import sys
import time

try:
    from urllib.request import urlopen
except ImportError:
    from urllib2 import urlopen

try:
    from concurrent.futures import ThreadPoolExecutor
except ImportError:
    ThreadPoolExecutor = lambda _: None  # noqa: E731
try:
    from concurrent.futures import ProcessPoolExecutor
except ImportError:
    ProcessPoolExecutor = lambda _: None  # noqa: E731

from mpi4py.futures import MPIPoolExecutor, as_completed

URLS = [
    "http://www.google.com/",
    "http://www.apple.com/",
    "http://www.ibm.com",
    "http://www.thisurlprobablydoesnotexist.com",
    "http://www.slashdot.org/",
    "http://www.python.org/",
    "http://www.bing.com/",
    "http://www.facebook.com/",
    "http://www.github.com/",
    "http://www.youtube.com/",
    "http://www.blogger.com/",
]


def load_url(url, timeout):
    return urlopen(url, timeout=timeout).read()


def download_urls_sequential(urls, timeout=60):
    url_to_content = {}
    for url in urls:
        with contextlib.suppress(Exception):
            url_to_content[url] = load_url(url, timeout=timeout)
    return url_to_content


def download_urls_with_executor(executor, urls, timeout=60):
    if executor is None:
        return {}
    try:
        url_to_content = {}
        future_to_url = {
            executor.submit(load_url, url, timeout): url for url in urls
        }
        for future in as_completed(future_to_url):
            with contextlib.suppress(Exception):
                url_to_content[future_to_url[future]] = future.result()
        return url_to_content
    finally:
        executor.shutdown()


def main():
    for meth, fn in [
        ("sequential", functools.partial(download_urls_sequential, URLS)),
        (
            "threads",
            functools.partial(
                download_urls_with_executor, ThreadPoolExecutor(10), URLS
            ),
        ),
        (
            "processes",
            functools.partial(
                download_urls_with_executor, ProcessPoolExecutor(10), URLS
            ),
        ),
        (
            "mpi4py",
            functools.partial(
                download_urls_with_executor, MPIPoolExecutor(10), URLS
            ),
        ),
    ]:
        sys.stdout.write(f"{meth.ljust(11)}: ")
        sys.stdout.flush()
        start = time.time()
        url_map = fn()
        elapsed = time.time() - start
        m, n = len(url_map), len(URLS)
        sys.stdout.write(
            f"{elapsed:5.2f} seconds ({m:2d} of {n:d} downloaded)\n",
        )
        sys.stdout.flush()


if __name__ == "__main__":
    main()



# Author:  Lisandro Dalcin
# Contact: dalcinl@gmail.com
"""Execute computations asynchronously using MPI processes."""
# pylint: disable=redefined-builtin

from ._base import (
    ALL_COMPLETED,
    FIRST_COMPLETED,
    FIRST_EXCEPTION,
    BrokenExecutor,
    CancelledError,
    Executor,
    Future,
    InvalidStateError,
    TimeoutError,  # noqa: A004
    as_completed,
    wait,
)
from .pool import (
    MPICommExecutor,
    MPIPoolExecutor,
    ProcessPoolExecutor,
    ThreadPoolExecutor,
    get_comm_workers,
)
from .util import (
    collect,
    compose,
)

__all__ = [
    "Future",
    "Executor",
    "wait",
    "FIRST_COMPLETED",
    "FIRST_EXCEPTION",
    "ALL_COMPLETED",
    "as_completed",
    "CancelledError",
    "TimeoutError",
    "InvalidStateError",
    "BrokenExecutor",
    "MPIPoolExecutor",
    "MPICommExecutor",
    "ThreadPoolExecutor",
    "ProcessPoolExecutor",
    "get_comm_workers",
    "collect",
    "compose",
]



#!/usr/bin/env python

if False:
    import mpi4py

    name = "name"  # lib{name}.so
    path = []
    mpi4py.profile(name, path=path)

# Import the MPI extension module
from mpi4py import MPI

if False:  # set to True to disable profiling
    MPI.Pcontrol(0)

# Import the 'array' module
from array import array

# This is just to make the logging
# output a bit more interesting
from time import sleep

comm = MPI.COMM_WORLD
nprocs = comm.Get_size()
myrank = comm.Get_rank()

n = array("i", [0])
pi = array("d", [0])
mypi = array("d", [0])


def comp_pi(n, myrank=0, nprocs=1):
    h = 1.0 / n
    s = 0.0
    for i in range(myrank + 1, n + 1, nprocs):
        x = h * (i - 0.5)
        s += 4.0 / (1.0 + x**2)
    return s * h


comm.Barrier()

for N in [10000] * 10:
    if myrank == 0:
        n[0] = N

    comm.Bcast([n, MPI.INT], root=0)

    mypi[0] = comp_pi(n[0], myrank, nprocs)

    comm.Reduce([mypi, MPI.DOUBLE], [pi, MPI.DOUBLE], op=MPI.SUM, root=0)

    comm.Barrier()

    sleep(0.01)



import contextlib

import arrayimpl
import mpiunittest as unittest

from mpi4py import MPI

scalar = arrayimpl.scalar


def mkzeros(n):
    return bytearray(n)


def memzero(m):
    try:
        m[:] = 0
    except IndexError:  # cffi buffer
        m[0 : len(m)] = b"\0" * len(m)


@contextlib.contextmanager
def win_lock(win, rank, *args, **kwargs):
    win.Lock(rank, *args, **kwargs)
    try:
        yield
    finally:
        win.Unlock(rank)


class BaseTestRMA:
    #
    COMM = MPI.COMM_NULL
    INFO = MPI.INFO_NULL

    COUNT_MIN = 0

    def setUp(self):
        nbytes = 100 * MPI.DOUBLE.size
        try:
            self.mpi_memory = MPI.Alloc_mem(nbytes)
            self.memory = self.mpi_memory
            memzero(self.memory)
        except MPI.Exception:
            self.mpi_memory = None
            self.memory = bytearray(nbytes)
        self.WIN = MPI.Win.Create(self.memory, 1, self.INFO, self.COMM)

    def tearDown(self):
        self.WIN.Free()
        if self.mpi_memory:
            MPI.Free_mem(self.mpi_memory)

    @unittest.skipMPI("impi(>=2021.13.0)")
    def testPutGet(self):
        group = self.WIN.Get_group()
        size = group.Get_size()
        group.Free()
        for array, typecode in arrayimpl.loop():
            with arrayimpl.test(self):
                if unittest.is_mpi_gpu("mvapich", array):
                    continue
                for count in range(10):
                    for rank in range(size):
                        with self.subTest(rank=rank, count=count):
                            sbuf = array([rank] * count, typecode)
                            rbuf = array(-1, typecode, count + 1)
                            self.WIN.Fence()
                            with win_lock(self.WIN, rank):
                                r = self.WIN.Rput(sbuf.as_mpi(), rank)
                                r.Wait()
                                self.WIN.Flush(rank)
                                r = self.WIN.Rget(rbuf.as_mpi_c(count), rank)
                                r.Wait()
                            for i in range(count):
                                self.assertEqual(sbuf[i], scalar(rank))
                                self.assertEqual(rbuf[i], scalar(rank))
                            self.assertEqual(rbuf[-1], scalar(-1))

    @unittest.skipMPI("impi(>=2021.13.0)")
    @unittest.skipMPI("openmpi(>=1.10.0,<1.11.0)")
    def testAccumulate(self):
        group = self.WIN.Get_group()
        size = group.Get_size()
        group.Free()
        for array, typecode in arrayimpl.loop():
            with arrayimpl.test(self):
                if unittest.is_mpi_gpu("openmpi", array):
                    continue
                if unittest.is_mpi_gpu("mvapich", array):
                    continue
                if typecode == "?":
                    continue
                if typecode in "FDG":
                    continue
                for count in range(10):
                    for rank in range(size):
                        with self.subTest(rank=rank, count=count):
                            ones = array([1] * count, typecode)
                            sbuf = array(range(count), typecode)
                            rbuf = array(-1, typecode, count + 1)
                            for op in (
                                MPI.SUM,
                                MPI.PROD,
                                MPI.MAX,
                                MPI.MIN,
                                MPI.REPLACE,
                            ):
                                with win_lock(self.WIN, rank):
                                    self.WIN.Put(ones.as_mpi(), rank)
                                    self.WIN.Flush(rank)
                                    r = self.WIN.Raccumulate(
                                        sbuf.as_mpi(), rank, op=op
                                    )
                                    r.Wait()
                                    self.WIN.Flush(rank)
                                    r = self.WIN.Rget(
                                        rbuf.as_mpi_c(count), rank
                                    )
                                    r.Wait()
                                #
                                for i in range(count):
                                    self.assertEqual(sbuf[i], scalar(i))
                                    self.assertEqual(rbuf[i], scalar(op(1, i)))
                                self.assertEqual(rbuf[-1], scalar(-1))

    @unittest.skipMPI("impi(>=2021.13.0)")
    @unittest.skipMPI("openmpi(>=1.10,<1.11)")
    def testGetAccumulate(self):
        group = self.WIN.Get_group()
        size = group.Get_size()
        group.Free()
        for array, typecode in arrayimpl.loop():
            with arrayimpl.test(self):
                if unittest.is_mpi_gpu("openmpi", array):
                    continue
                if unittest.is_mpi_gpu("mvapich", array):
                    continue
                if typecode == "?":
                    continue
                if typecode in "FDG":
                    continue
                for count in range(10):
                    for rank in range(size):
                        with self.subTest(rank=rank, count=count):
                            ones = array([1] * count, typecode)
                            sbuf = array(range(count), typecode)
                            rbuf = array(-1, typecode, count + 1)
                            gbuf = array(-1, typecode, count + 1)
                            for op in (
                                MPI.SUM,
                                MPI.PROD,
                                MPI.MAX,
                                MPI.MIN,
                                MPI.REPLACE,
                                MPI.NO_OP,
                            ):
                                with win_lock(self.WIN, rank):
                                    self.WIN.Put(ones.as_mpi(), rank)
                                    self.WIN.Flush(rank)
                                    r = self.WIN.Rget_accumulate(
                                        sbuf.as_mpi(),
                                        rbuf.as_mpi_c(count),
                                        rank,
                                        op=op,
                                    )
                                    r.Wait()
                                    self.WIN.Flush(rank)
                                    r = self.WIN.Rget(
                                        gbuf.as_mpi_c(count), rank
                                    )
                                    r.Wait()
                                #
                                for i in range(count):
                                    self.assertEqual(sbuf[i], scalar(i))
                                    self.assertEqual(rbuf[i], scalar(1))
                                    self.assertEqual(gbuf[i], scalar(op(1, i)))
                                self.assertEqual(rbuf[-1], scalar(-1))
                                self.assertEqual(gbuf[-1], scalar(-1))

    def testPutProcNull(self):
        rank = self.COMM.Get_rank()
        with win_lock(self.WIN, rank):
            r = self.WIN.Rput(None, MPI.PROC_NULL, None)
            r.Wait()

    def testGetProcNull(self):
        rank = self.COMM.Get_rank()
        with win_lock(self.WIN, rank):
            r = self.WIN.Rget(None, MPI.PROC_NULL, None)
            r.Wait()

    def testAccumulateProcNullReplace(self):
        rank = self.COMM.Get_rank()
        zeros = mkzeros(8)
        with win_lock(self.WIN, rank):
            r = self.WIN.Raccumulate(
                [zeros, MPI.INT], MPI.PROC_NULL, None, MPI.REPLACE
            )
            r.Wait()
            r = self.WIN.Raccumulate(
                [zeros, MPI.INT], MPI.PROC_NULL, None, MPI.REPLACE
            )
            r.Wait()

    def testAccumulateProcNullSum(self):
        rank = self.COMM.Get_rank()
        zeros = mkzeros(8)
        with win_lock(self.WIN, rank):
            r = self.WIN.Raccumulate(
                [zeros, MPI.INT], MPI.PROC_NULL, None, MPI.SUM
            )
            r.Wait()
            r = self.WIN.Raccumulate(
                [None, MPI.INT], MPI.PROC_NULL, None, MPI.SUM
            )
            r.Wait()


@unittest.skipMPI("MPI(<3.0)")
@unittest.skipMPI("openmpi(<1.8.1)")
@unittest.skipMPI("MPICH2(<1.5.0)")
class TestRMASelf(BaseTestRMA, unittest.TestCase):
    #
    COMM = MPI.COMM_SELF


@unittest.skipMPI("MPI(<3.0)")
@unittest.skipMPI("openmpi(<1.8.1)")
@unittest.skipMPI("MPICH2(<1.5.0)")
class TestRMAWorld(BaseTestRMA, unittest.TestCase):
    #
    COMM = MPI.COMM_WORLD


try:
    MPI.Win.Create(None, 1, MPI.INFO_NULL, MPI.COMM_SELF).Free()
except (NotImplementedError, MPI.Exception):
    unittest.disable(BaseTestRMA, "mpi-rma-nb")


if __name__ == "__main__":
    unittest.main()



import os

from mpi4py import rc

assert rc.initialize is True
assert rc.finalize is None
assert rc.threads is True
assert rc.thread_level == "multiple"

os.environ["MPI4PY_RC_INITIALIZE"] = "false"
os.environ["MPI4PY_RC_FINALIZE"] = "true"
os.environ["MPI4PY_RC_THREADS"] = "false"
os.environ["MPI4PY_RC_THREAD_LEVEL"] = "single"

os.environ["MPI4PY_RC_IRECV_BUFSZ"] = str(2048)
os.environ["MPI4PY_PICKLE_PROTOCOL"] = str(3)
os.environ["MPI4PY_PICKLE_THRESHOLD"] = str(1024)

from mpi4py import MPI  # noqa: E402

assert not MPI.Is_initialized()
assert not MPI.Is_finalized()

assert rc.initialize is False
assert rc.finalize is True
assert rc.threads is False
assert rc.thread_level == "single"

assert rc.irecv_bufsz == 2048
assert MPI.pickle.PROTOCOL == 3
assert MPI.pickle.THRESHOLD == 1024



from libmpi import ffi, lib

NULL = ffi.NULL
size_p = ffi.new("int*")
rank_p = ffi.new("int*")
nlen_p = ffi.new("int*")
name_p = ffi.new("char[]", lib.MPI_MAX_PROCESSOR_NAME)
lib.MPI_Init(NULL, NULL)
lib.MPI_Comm_size(lib.MPI_COMM_WORLD, size_p)
lib.MPI_Comm_rank(lib.MPI_COMM_WORLD, rank_p)
lib.MPI_Get_processor_name(name_p, nlen_p)

size = size_p[0]
rank = rank_p[0]
nlen = nlen_p[0]
name = ffi.string(name_p[0:nlen]).decode()

print(f"Hello, World! I am process {rank} of {size} on {name}.")

lib.MPI_Finalize()



# -----------------------------------------------------------------------------

cdef extern from "Python.h":
    object PyOS_FSPath(object)
    object PyUnicode_EncodeFSDefault(object)


cdef inline object asmpifspath(object path, char *p[]):
    path = PyOS_FSPath(path)
    if PyUnicode_Check(path):
        path = PyUnicode_EncodeFSDefault(path)
    PyBytes_AsStringAndSize(path, p, NULL)
    return path

# -----------------------------------------------------------------------------



# Create flavors
# --------------
WIN_FLAVOR_CREATE   = MPI_WIN_FLAVOR_CREATE
WIN_FLAVOR_ALLOCATE = MPI_WIN_FLAVOR_ALLOCATE
WIN_FLAVOR_DYNAMIC  = MPI_WIN_FLAVOR_DYNAMIC
WIN_FLAVOR_SHARED   = MPI_WIN_FLAVOR_SHARED

# Memory model
# ------------
WIN_SEPARATE = MPI_WIN_SEPARATE
WIN_UNIFIED  = MPI_WIN_UNIFIED

# Assertion modes
# ---------------
MODE_NOCHECK   = MPI_MODE_NOCHECK
MODE_NOSTORE   = MPI_MODE_NOSTORE
MODE_NOPUT     = MPI_MODE_NOPUT
MODE_NOPRECEDE = MPI_MODE_NOPRECEDE
MODE_NOSUCCEED = MPI_MODE_NOSUCCEED

# Lock types
# ----------
LOCK_EXCLUSIVE = MPI_LOCK_EXCLUSIVE
LOCK_SHARED    = MPI_LOCK_SHARED


cdef class Win:
    """
    Remote memory access context.
    """

    def __cinit__(self, Win win: Win | None = None):
        cinit(self, win)

    def __dealloc__(self):
        dealloc(self)

    def __richcmp__(self, other, int op):
        if not isinstance(other, Win): return NotImplemented
        return richcmp(self, other, op)

    def __bool__(self) -> bool:
        return nonnull(self)

    def __reduce__(self) -> str | tuple[Any, ...]:
        return reduce_default(self)

    property handle:
        """MPI handle."""
        def __get__(self) -> int:
            return tohandle(self)

    @classmethod
    def fromhandle(cls, handle: int) -> Win:
        """
        Create object from MPI handle.
        """
        return fromhandle(<MPI_Win> <Py_uintptr_t> handle)

    def free(self) -> None:
        """
        Call `Free` if not null.
        """
        safefree(self)

    # Window Creation
    # ---------------

    @classmethod
    def Create(
        cls,
        memory: Buffer | Bottom,
        Aint disp_unit: int = 1,
        Info info: Info = INFO_NULL,
        Intracomm comm: Intracomm = COMM_SELF,
    ) -> Self:
        """
        Create an window object for one-sided communication.
        """
        cdef void *base = MPI_BOTTOM
        cdef MPI_Aint size = 0
        if is_BOTTOM(memory):
            memory = None
        if memory is not None:
            memory = asbuffer_w(memory, &base, &size)
        cdef Win win = <Win>New(cls)
        with nogil: CHKERR( MPI_Win_create_c(
            base, size, disp_unit,
            info.ob_mpi, comm.ob_mpi, &win.ob_mpi) )
        options_set_errhandler(win.ob_mpi)
        win.ob_mem = memory
        return win

    @classmethod
    def Allocate(
        cls,
        Aint size: int,
        Aint disp_unit: int = 1,
        Info info: Info = INFO_NULL,
        Intracomm comm: Intracomm = COMM_SELF,
    ) -> Self:
        """
        Create an window object for one-sided communication.
        """
        cdef void *base = NULL
        cdef Win win = <Win>New(cls)
        with nogil: CHKERR( MPI_Win_allocate_c(
            size, disp_unit, info.ob_mpi,
            comm.ob_mpi, &base, &win.ob_mpi) )
        options_set_errhandler(win.ob_mpi)
        return win

    @classmethod
    def Allocate_shared(
        cls,
        Aint size: int,
        Aint disp_unit: int = 1,
        Info info: Info = INFO_NULL,
        Intracomm comm: Intracomm = COMM_SELF,
    ) -> Self:
        """
        Create an window object for one-sided communication.
        """
        cdef void *base = NULL
        cdef Win win = <Win>New(cls)
        with nogil: CHKERR( MPI_Win_allocate_shared_c(
            size, disp_unit, info.ob_mpi,
            comm.ob_mpi, &base, &win.ob_mpi) )
        options_set_errhandler(win.ob_mpi)
        return win

    def Shared_query(self, int rank: int) -> tuple[buffer, int]:
        """
        Query the process-local address for remote memory segments.
        """
        cdef void *base = NULL
        cdef MPI_Aint size = 0
        cdef MPI_Aint disp_unit = 1
        with nogil: CHKERR( MPI_Win_shared_query_c(
            self.ob_mpi, rank,
            &size, &disp_unit, &base) )
        return (tobuffer(self, base, size, 0), disp_unit)

    @classmethod
    def Create_dynamic(
        cls,
        Info info: Info = INFO_NULL,
        Intracomm comm: Intracomm = COMM_SELF,
    ) -> Self:
        """
        Create an window object for one-sided communication.
        """
        cdef Win win = <Win>New(cls)
        with nogil: CHKERR( MPI_Win_create_dynamic(
            info.ob_mpi, comm.ob_mpi, &win.ob_mpi) )
        options_set_errhandler(win.ob_mpi)
        win.ob_mem = {}
        return win

    def Attach(self, memory: Buffer) -> None:
        """
        Attach a local memory region.
        """
        cdef void *base = NULL
        cdef MPI_Aint size = 0
        memory = asbuffer_w(memory, &base, &size)
        with nogil: CHKERR( MPI_Win_attach(self.ob_mpi, base, size) )
        try:
            if self.ob_mem is None: self.ob_mem = {}
            (<dict>self.ob_mem)[<MPI_Aint>base] = memory
        except:   # ~> uncovered  # noqa
            pass  # ~> uncovered

    def Detach(self, memory: Buffer) -> None:
        """
        Detach a local memory region.
        """
        cdef void *base = NULL
        memory = asbuffer_w(memory, &base, NULL)
        with nogil: CHKERR( MPI_Win_detach(self.ob_mpi, base) )
        try:
            if self.ob_mem is None: return
            del (<dict>self.ob_mem)[<MPI_Aint>base]
        except:   # ~> uncovered  # noqa
            pass  # ~> uncovered

    def Free(self) -> None:
        """
        Free a window.
        """
        cdef MPI_Win save = self.ob_mpi
        with nogil: CHKERR( MPI_Win_free(&self.ob_mpi) )
        if constobj(self): self.ob_mpi = save
        self.ob_mem = None

    # Window Info
    # -----------

    def Set_info(self, Info info: Info) -> None:
        """
        Set new values for the hints associated with a window.
        """
        with nogil: CHKERR( MPI_Win_set_info(self.ob_mpi, info.ob_mpi) )

    def Get_info(self) -> Info:
        """
        Return the current hints for a window.
        """
        cdef Info info = <Info>New(Info)
        with nogil: CHKERR( MPI_Win_get_info( self.ob_mpi, &info.ob_mpi) )
        return info

    property info:
        """Info hints."""
        def __get__(self) -> Info:
            return self.Get_info()

        def __set__(self, value: Info):
            self.Set_info(value)

    # Window Group
    # -------------

    def Get_group(self) -> Group:
        """
        Access the group of processes that created the window.
        """
        cdef Group group = Group()
        with nogil: CHKERR( MPI_Win_get_group(self.ob_mpi, &group.ob_mpi) )
        return group

    property group:
        """Group."""
        def __get__(self) -> Group:
            return self.Get_group()

    property group_size:
        """Group size."""
        def __get__(self) -> int:
            cdef MPI_Group group = MPI_GROUP_NULL
            cdef int group_size  = -1
            CHKERR( MPI_Win_get_group(self.ob_mpi, &group) )
            try:
                CHKERR( MPI_Group_size(group, &group_size) )
            finally:
                CHKERR( MPI_Group_free(&group) )
            return group_size

    property group_rank:
        """Group rank."""
        def __get__(self) -> int:
            cdef MPI_Group group = MPI_GROUP_NULL
            cdef int group_rank  = MPI_PROC_NULL
            CHKERR( MPI_Win_get_group(self.ob_mpi, &group) )
            try:
                CHKERR( MPI_Group_rank(group, &group_rank) )
            finally:
                CHKERR( MPI_Group_free(&group) )
            return group_rank

    # Window Attributes
    # -----------------

    def Get_attr(self, int keyval: int) -> int | Any | None:
        """
        Retrieve attribute value by key.
        """
        cdef void *attrval = NULL
        cdef int flag = 0
        CHKERR( MPI_Win_get_attr(self.ob_mpi, keyval, &attrval, &flag) )
        if flag == 0: return None
        if attrval == NULL: return 0
        # MPI-2 predefined attribute keyvals
        if keyval == MPI_WIN_BASE:
            return <MPI_Aint>attrval
        elif keyval == MPI_WIN_SIZE:
            return (<MPI_Aint*>attrval)[0]
        elif keyval == MPI_WIN_DISP_UNIT:
            return (<int*>attrval)[0]
        # MPI-3 predefined attribute keyvals
        elif keyval == MPI_WIN_CREATE_FLAVOR:
            return (<int*>attrval)[0]
        elif keyval == MPI_WIN_MODEL:
            return (<int*>attrval)[0]
        # user-defined attribute keyval
        return PyMPI_attr_get(self.ob_mpi, keyval, attrval)

    def Set_attr(self, int keyval: int, attrval: Any) -> None:
        """
        Store attribute value associated with a key.
        """
        PyMPI_attr_set(self.ob_mpi, keyval, attrval)

    def Delete_attr(self, int keyval: int) -> None:
        """
        Delete attribute value associated with a key.
        """
        PyMPI_attr_del(self.ob_mpi, keyval)

    @classmethod
    def Create_keyval(
        cls,
        copy_fn: Callable[[Win, int, Any], Any] | None = None,
        delete_fn: Callable[[Win, int, Any], None] | None = None,
        nopython: bool = False,
    ) -> int:
        """
        Create a new attribute key for windows.
        """
        cdef int keyval = MPI_KEYVAL_INVALID
        cdef MPI_Win_copy_attr_function *_copy = PyMPI_attr_copy_fn
        cdef MPI_Win_delete_attr_function *_del = PyMPI_attr_delete_fn
        cdef _p_keyval state = _p_keyval(copy_fn, delete_fn, nopython)
        CHKERR( MPI_Win_create_keyval(_copy, _del, &keyval, <void *>state) )
        PyMPI_attr_state_set(MPI_WIN_NULL, keyval, state)
        return keyval

    @classmethod
    def Free_keyval(cls, int keyval: int) -> int:
        """
        Free an attribute key for windows.
        """
        cdef int keyval_save = keyval
        CHKERR( MPI_Win_free_keyval(&keyval) )
        PyMPI_attr_state_del(MPI_WIN_NULL, keyval_save)
        return keyval

    property attrs:
        "Attributes."
        def __get__(self) -> tuple[int, int, int]:
            cdef void *base = NULL
            cdef MPI_Aint size = 0
            cdef int disp_unit = 1
            win_get_base(self.ob_mpi, &base)
            win_get_size(self.ob_mpi, &size)
            win_get_unit(self.ob_mpi, &disp_unit)
            return (<MPI_Aint>base, size, disp_unit)

    property flavor:
        """Create flavor."""
        def __get__(self) -> int:
            cdef int keyval = MPI_WIN_CREATE_FLAVOR
            cdef int *attrval = NULL, flag = 0
            cdef int flavor = MPI_WIN_FLAVOR_CREATE
            if keyval != MPI_KEYVAL_INVALID:
                CHKERR( MPI_Win_get_attr(self.ob_mpi, keyval,
                                         <void*>&attrval, &flag) )
                if flag and attrval != NULL:
                    flavor = attrval[0]
            return flavor

    property model:
        """Memory model."""
        def __get__(self) -> int:
            cdef int keyval = MPI_WIN_MODEL
            cdef int *attrval = NULL, flag = 0
            cdef int model = MPI_WIN_SEPARATE
            if keyval != MPI_KEYVAL_INVALID:
                CHKERR( MPI_Win_get_attr(self.ob_mpi, keyval,
                                         <void*>&attrval, &flag) )
                if flag and attrval != NULL:
                    model = attrval[0]
            return model

    def tomemory(self) -> buffer:
        """
        Return window memory buffer.
        """
        return getbuffer(self, 0, 1)

    # buffer interface (PEP 3118)

    def __getbuffer__(self, Py_buffer *view, int flags):
        cdef void *base = NULL
        cdef MPI_Aint size = 0
        win_get_base(self.ob_mpi, &base)
        win_get_size(self.ob_mpi, &size)
        PyBuffer_FillInfo(view, self, base, size, 0, flags)

    # Communication Operations
    # ------------------------

    def Put(
        self,
        origin: BufSpec,
        int target_rank: int,
        target: TargetSpec | None = None,
    ) -> None:
        """
        Put data into a memory window on a remote process.
        """
        cdef _p_msg_rma msg = message_rma()
        msg.for_put(origin, target_rank, target)
        with nogil: CHKERR( MPI_Put_c(
            msg.oaddr, msg.ocount, msg.otype,
            target_rank,
            msg.tdisp, msg.tcount, msg.ttype,
            self.ob_mpi) )

    def Get(
        self,
        origin: BufSpec,
        int target_rank: int,
        target: TargetSpec | None = None,
    ) -> None:
        """
        Get data from a memory window on a remote process.
        """
        cdef _p_msg_rma msg = message_rma()
        msg.for_get(origin, target_rank, target)
        with nogil: CHKERR( MPI_Get_c(
            msg.oaddr, msg.ocount, msg.otype,
            target_rank,
            msg.tdisp, msg.tcount, msg.ttype,
            self.ob_mpi) )

    def Accumulate(
        self,
        origin: BufSpec,
        int target_rank: int,
        target: TargetSpec | None = None,
        Op op: Op = SUM,
    ) -> None:
        """
        Accumulate data into the target process.
        """
        cdef _p_msg_rma msg = message_rma()
        msg.for_acc(origin, target_rank, target)
        with nogil: CHKERR( MPI_Accumulate_c(
            msg.oaddr, msg.ocount, msg.otype,
            target_rank,
            msg.tdisp, msg.tcount, msg.ttype,
            op.ob_mpi, self.ob_mpi) )

    def Get_accumulate(
        self,
        origin: BufSpec,
        result: BufSpec,
        int target_rank: int,
        target: TargetSpec | None = None,
        Op op: Op = SUM,
    ) -> None:
        """
        Fetch-and-accumulate data into the target process.
        """
        cdef _p_msg_rma msg = message_rma()
        msg.for_get_acc(origin, result, target_rank, target)
        with nogil: CHKERR( MPI_Get_accumulate_c(
            msg.oaddr, msg.ocount, msg.otype,
            msg.raddr, msg.rcount, msg.rtype,
            target_rank,
            msg.tdisp, msg.tcount, msg.ttype,
            op.ob_mpi, self.ob_mpi) )

    def Fetch_and_op(
        self,
        origin: BufSpec,
        result: BufSpec,
        int target_rank: int,
        Aint target_disp: int = 0,
        Op op: Op = SUM,
    ) -> None:
        """
        Perform one-sided read-modify-write.
        """
        cdef _p_msg_rma msg = message_rma()
        msg.for_fetch_op(origin, result, target_rank, target_disp)
        with nogil: CHKERR( MPI_Fetch_and_op(
                msg.oaddr, msg.raddr, msg.ttype,
                target_rank, target_disp,
                op.ob_mpi, self.ob_mpi) )

    def Compare_and_swap(
        self,
        origin: BufSpec,
        compare: BufSpec,
        result: BufSpec,
        int target_rank: int,
        Aint target_disp: int = 0,
    ) -> None:
        """
        Perform one-sided atomic compare-and-swap.
        """
        cdef _p_msg_rma msg = message_rma()
        msg.for_cmp_swap(origin, compare, result, target_rank, target_disp)
        with nogil: CHKERR( MPI_Compare_and_swap(
                msg.oaddr, msg.caddr, msg.raddr, msg.ttype,
                target_rank, target_disp, self.ob_mpi) )

    # Request-based RMA Communication Operations
    # ------------------------------------------

    def Rput(
        self,
        origin: BufSpec,
        int target_rank: int,
        target: TargetSpec | None = None,
    ) -> Request:
        """
        Put data into a memory window on a remote process.
        """
        cdef _p_msg_rma msg = message_rma()
        msg.for_put(origin, target_rank, target)
        cdef Request request = <Request>New(Request)
        with nogil: CHKERR( MPI_Rput_c(
            msg.oaddr, msg.ocount, msg.otype,
            target_rank,
            msg.tdisp, msg.tcount, msg.ttype,
            self.ob_mpi, &request.ob_mpi) )
        request.ob_buf = msg
        return request

    def Rget(
        self,
        origin: BufSpec,
        int target_rank: int,
        target: TargetSpec | None = None,
    ) -> Request:
        """
        Get data from a memory window on a remote process.
        """
        cdef _p_msg_rma msg = message_rma()
        msg.for_get(origin, target_rank, target)
        cdef Request request = <Request>New(Request)
        with nogil: CHKERR( MPI_Rget_c(
            msg.oaddr, msg.ocount, msg.otype,
            target_rank,
            msg.tdisp, msg.tcount, msg.ttype,
            self.ob_mpi, &request.ob_mpi) )
        request.ob_buf = msg
        return request

    def Raccumulate(
        self,
        origin: BufSpec,
        int target_rank: int,
        target: TargetSpec | None = None,
        Op op: Op = SUM,
    ) -> Request:
        """
        Fetch-and-accumulate data into the target process.
        """
        cdef _p_msg_rma msg = message_rma()
        msg.for_acc(origin, target_rank, target)
        cdef Request request = <Request>New(Request)
        with nogil: CHKERR( MPI_Raccumulate_c(
            msg.oaddr, msg.ocount, msg.otype,
            target_rank,
            msg.tdisp, msg.tcount, msg.ttype,
            op.ob_mpi, self.ob_mpi, &request.ob_mpi) )
        request.ob_buf = msg
        return request

    def Rget_accumulate(
        self,
        origin: BufSpec,
        result: BufSpec,
        int target_rank: int,
        target: TargetSpec | None = None,
        Op op: Op = SUM,
    ) -> Request:
        """
        Accumulate data into the target process using remote memory access.
        """
        cdef _p_msg_rma msg = message_rma()
        msg.for_get_acc(origin, result, target_rank, target)
        cdef Request request = <Request>New(Request)
        with nogil: CHKERR( MPI_Rget_accumulate_c(
            msg.oaddr, msg.ocount, msg.otype,
            msg.raddr, msg.rcount, msg.rtype,
            target_rank,
            msg.tdisp, msg.tcount, msg.ttype,
            op.ob_mpi, self.ob_mpi, &request.ob_mpi) )
        request.ob_buf = msg
        return request

    # Synchronization Calls
    # ---------------------

    # Fence
    # -----

    def Fence(self, int assertion: int = 0) -> None:
        """
        Perform an MPI fence synchronization on a window.
        """
        with nogil: CHKERR( MPI_Win_fence(assertion, self.ob_mpi) )

    # General Active Target Synchronization
    # -------------------------------------

    def Start(self, Group group: Group, int assertion: int = 0) -> None:
        """
        Start an RMA access epoch for MPI.
        """
        with nogil: CHKERR( MPI_Win_start(
            group.ob_mpi, assertion, self.ob_mpi) )

    def Complete(self) -> None:
        """
        Complete an RMA operation begun after an `Start`.
        """
        with nogil: CHKERR( MPI_Win_complete(self.ob_mpi) )

    def Post(self, Group group: Group, int assertion: int = 0) -> None:
        """
        Start an RMA exposure epoch.
        """
        with nogil: CHKERR( MPI_Win_post(
            group.ob_mpi, assertion, self.ob_mpi) )

    def Wait(self) -> Literal[True]:
        """
        Complete an RMA exposure epoch begun with `Post`.
        """
        with nogil: CHKERR( MPI_Win_wait(self.ob_mpi) )
        return True

    def Test(self) -> bool:
        """
        Test whether an RMA exposure epoch has completed.
        """
        cdef int flag = 0
        with nogil: CHKERR( MPI_Win_test(self.ob_mpi, &flag) )
        return <bint>flag

    # Lock
    # ----

    def Lock(
        self,
        int rank: int,
        int lock_type: int = LOCK_EXCLUSIVE,
        int assertion: int = 0,
    ) -> None:
        """
        Begin an RMA access epoch at the target process.
        """
        with nogil: CHKERR( MPI_Win_lock(
            lock_type, rank, assertion, self.ob_mpi) )

    def Unlock(self, int rank: int) -> None:
        """
        Complete an RMA access epoch at the target process.
        """
        with nogil: CHKERR( MPI_Win_unlock(rank, self.ob_mpi) )

    def Lock_all(self, int assertion: int = 0) -> None:
        """
        Begin an RMA access epoch at all processes.
        """
        with nogil: CHKERR( MPI_Win_lock_all(assertion, self.ob_mpi) )

    def Unlock_all(self) -> None:
        """
        Complete an RMA access epoch at all processes.
        """
        with nogil: CHKERR( MPI_Win_unlock_all(self.ob_mpi) )

    # Flush and Sync
    # --------------

    def Flush(self, int rank: int) -> None:
        """
        Complete all outstanding RMA operations at a target.
        """
        with nogil: CHKERR( MPI_Win_flush(rank, self.ob_mpi) )

    def Flush_all(self) -> None:
        """
        Complete all outstanding RMA operations at all targets.
        """
        with nogil: CHKERR( MPI_Win_flush_all(self.ob_mpi) )

    def Flush_local(self, int rank: int) -> None:
        """
        Complete locally all outstanding RMA operations at a target.
        """
        with nogil: CHKERR( MPI_Win_flush_local(rank, self.ob_mpi) )

    def Flush_local_all(self) -> None:
        """
        Complete locally all outstanding RMA operations at all targets.
        """
        with nogil: CHKERR( MPI_Win_flush_local_all(self.ob_mpi) )

    def Sync(self) -> None:
        """
        Synchronize public and private copies of the window.
        """
        with nogil: CHKERR( MPI_Win_sync(self.ob_mpi) )

    # Error Handling
    # --------------

    @classmethod
    def Create_errhandler(
        cls,
        errhandler_fn: Callable[[Win, int], None],
    ) -> Errhandler:
        """
        Create a new error handler for windows.
        """
        cdef Errhandler errhandler = <Errhandler>New(Errhandler)
        cdef MPI_Win_errhandler_function *fn = NULL
        cdef int index = errhdl_new(errhandler_fn, &fn)
        try:
            CHKERR( MPI_Win_create_errhandler(fn, &errhandler.ob_mpi) )
        except:                     # ~> uncovered  # noqa
            errhdl_del(&index, fn)  # ~> uncovered
            raise                   # ~> uncovered
        return errhandler

    def Get_errhandler(self) -> Errhandler:
        """
        Get the error handler for a window.
        """
        cdef Errhandler errhandler = <Errhandler>New(Errhandler)
        CHKERR( MPI_Win_get_errhandler(self.ob_mpi, &errhandler.ob_mpi) )
        return errhandler

    def Set_errhandler(self, Errhandler errhandler: Errhandler) -> None:
        """
        Set the error handler for a window.
        """
        CHKERR( MPI_Win_set_errhandler(self.ob_mpi, errhandler.ob_mpi) )

    def Call_errhandler(self, int errorcode: int) -> None:
        """
        Call the error handler installed on a window.
        """
        CHKERR( MPI_Win_call_errhandler(self.ob_mpi, errorcode) )

    # Naming Objects
    # --------------

    def Get_name(self) -> str:
        """
        Get the print name for this window.
        """
        cdef char name[MPI_MAX_OBJECT_NAME+1]
        cdef int nlen = 0
        CHKERR( MPI_Win_get_name(self.ob_mpi, name, &nlen) )
        return tompistr(name, nlen)

    def Set_name(self, name: str) -> None:
        """
        Set the print name for this window.
        """
        cdef char *cname = NULL
        name = asmpistr(name, &cname)
        CHKERR( MPI_Win_set_name(self.ob_mpi, cname) )

    property name:
        """Print name."""
        def __get__(self) -> str:
            return self.Get_name()

        def __set__(self, value: str):
            self.Set_name(value)

    # Integer Handle
    # --------------

    def toint(self) -> int:
        """
        Translate object to integer handle.
        """
        return MPI_Win_toint(self.ob_mpi)

    @classmethod
    def fromint(cls, arg: int, /) -> Win:
        """
        Translate integer handle to object.
        """
        return fromhandle(MPI_Win_fromint(arg))

    # Fortran Handle
    # --------------

    def py2f(self) -> int:
        """
        """
        return MPI_Win_c2f(self.ob_mpi)

    @classmethod
    def f2py(cls, arg: int) -> Win:
        """
        """
        return fromhandle(MPI_Win_f2c(arg))


cdef Win __WIN_NULL__ = def_Win( MPI_WIN_NULL , "WIN_NULL" )


# Predefined window handles
# -------------------------

WIN_NULL = __WIN_NULL__  #: Null window handle



import mpiunittest as unittest

from mpi4py import MPI


class TestStatus(unittest.TestCase):
    #
    def setUp(self):
        self.STATUS = MPI.Status()

    def tearDown(self):
        self.STATUS = None

    def testDefaultFieldValues(self):
        self.assertEqual(self.STATUS.Get_source(), MPI.ANY_SOURCE)
        self.assertEqual(self.STATUS.Get_tag(), MPI.ANY_TAG)
        self.assertEqual(self.STATUS.Get_error(), MPI.SUCCESS)

    def testGetCount(self):
        count = self.STATUS.Get_count(MPI.BYTE)
        self.assertEqual(count, 0)

    def testGetElements(self):
        elements = self.STATUS.Get_elements(MPI.BYTE)
        self.assertEqual(elements, 0)

    def testSetElements(self):
        with self.catchNotImplementedError(2, 0):
            self.STATUS.Set_elements(MPI.BYTE, 7)
            count = self.STATUS.Get_count(MPI.BYTE)
            self.assertEqual(count, 7)
            elements = self.STATUS.Get_elements(MPI.BYTE)
            self.assertEqual(elements, 7)

    def testIsCancelled(self):
        flag = self.STATUS.Is_cancelled()
        self.assertIs(type(flag), bool)
        self.assertFalse(flag)

    def testSetCancelled(self):
        with self.catchNotImplementedError(2, 0):
            self.STATUS.Set_cancelled(True)
            flag = self.STATUS.Is_cancelled()
            self.assertTrue(flag)

    def testPyProps(self):
        self.assertEqual(self.STATUS.Get_source(), self.STATUS.source)
        self.assertEqual(self.STATUS.Get_tag(), self.STATUS.tag)
        self.assertEqual(self.STATUS.Get_error(), self.STATUS.error)
        self.STATUS.source = 1
        self.STATUS.tag = 2
        self.STATUS.error = MPI.ERR_ARG
        self.assertEqual(self.STATUS.source, 1)
        self.assertEqual(self.STATUS.tag, 2)
        self.assertEqual(self.STATUS.error, MPI.ERR_ARG)
        with self.catchNotImplementedError(2, 0):
            self.assertIs(type(self.STATUS.count), int)
            self.assertEqual(self.STATUS.count, 0)
            self.STATUS.count = 7
            self.assertEqual(self.STATUS.count, 7)
            self.STATUS.count = 0
        with self.catchNotImplementedError(2, 0):
            self.assertIs(type(self.STATUS.cancelled), bool)
            self.assertFalse(self.STATUS.cancelled)
            self.STATUS.cancelled = True
            self.assertTrue(self.STATUS.cancelled)
            self.STATUS.cancelled = False

    def testConstructor(self):
        self.assertRaises(TypeError, MPI.Status, 123)
        self.assertRaises(TypeError, MPI.Status, "abc")

    def testCopyConstructor(self):
        self.STATUS.source = 1
        self.STATUS.tag = 2
        self.STATUS.error = MPI.ERR_ARG
        status = MPI.Status(self.STATUS)
        self.assertEqual(status.source, 1)
        self.assertEqual(status.tag, 2)
        self.assertEqual(status.error, MPI.ERR_ARG)
        with self.catchNotImplementedError(2, 0):
            self.STATUS.Set_elements(MPI.BYTE, 7)
        with self.catchNotImplementedError(2, 0):
            self.STATUS.Set_cancelled(True)
        status = MPI.Status(self.STATUS)
        with self.catchNotImplementedError(2, 0):
            count = status.Get_count(MPI.BYTE)
            elems = status.Get_elements(MPI.BYTE)
            self.assertEqual(count, 7)
            self.assertEqual(elems, 7)
        with self.catchNotImplementedError(2, 0):
            flag = status.Is_cancelled()
            self.assertTrue(flag)

    def testPickle(self):
        from pickle import dumps, loads

        self.STATUS.source = 1
        self.STATUS.tag = 2
        self.STATUS.error = MPI.ERR_ARG
        status = loads(dumps(self.STATUS))
        self.assertEqual(status.source, 1)
        self.assertEqual(status.tag, 2)
        self.assertEqual(status.error, MPI.ERR_ARG)

    def testToMemory(self):
        status = self.STATUS
        status.source = 11
        status.tag = 22
        status.error = 33
        mem = status.tomemory()
        seq = list(mem)
        mem[seq.index(11)] = 111
        mem[seq.index(22)] = 222
        mem[seq.index(33)] = 333
        self.assertEqual(status.source, 111)
        self.assertEqual(status.tag, 222)
        self.assertEqual(status.error, 333)


if __name__ == "__main__":
    unittest.main()



import operator
from functools import reduce

import mpiunittest as unittest

from mpi4py import MPI


def cumsum(seq):
    return reduce(operator.add, seq, 0)


def cumprod(seq):
    return reduce(operator.mul, seq, 1)


_basic = [
    None,
    True,
    False,
    -7,
    0,
    7,
    2**31,
    -(2**63) + 1,
    2**63 - 1,
    -2.17,
    0.0,
    +1.62,
    1 + 2j,
    2 - 3j,
    "mpi4py",
]
messages = list(_basic)
messages += [
    list(_basic),
    tuple(_basic),
    {f"k{k}": v for k, v in enumerate(_basic)},
]


class BaseTestCCOObj:
    #
    COMM = MPI.COMM_NULL

    def testBarrier(self):
        self.COMM.barrier()

    def testBcast(self):
        for smess in messages:
            for root in range(self.COMM.Get_size()):
                rmess = self.COMM.bcast(smess, root=root)
                self.assertEqual(smess, rmess)

    def testGather(self):
        size = self.COMM.Get_size()
        rank = self.COMM.Get_rank()
        for smess in [*messages, messages]:
            for root in range(size):
                rmess = self.COMM.gather(smess, root=root)
                if rank == root:
                    self.assertEqual(rmess, [smess] * size)
                else:
                    self.assertIsNone(rmess)

    def testScatter(self):
        size = self.COMM.Get_size()
        rank = self.COMM.Get_rank()
        for smess in [*messages, messages]:
            for root in range(size):
                if rank == root:
                    rmess = self.COMM.scatter([smess] * size, root=root)
                else:
                    rmess = self.COMM.scatter(None, root=root)
                self.assertEqual(rmess, smess)

    def testAllgather(self):
        size = self.COMM.Get_size()
        self.COMM.Get_rank()
        for smess in [*messages, messages]:
            rmess = self.COMM.allgather(smess)
            self.assertEqual(rmess, [smess] * size)

    def testAlltoall(self):
        size = self.COMM.Get_size()
        self.COMM.Get_rank()
        for smess in [*messages, messages]:
            rmess = self.COMM.alltoall([smess] * size)
            self.assertEqual(rmess, [smess] * size)

    def testReduce(self):
        size = self.COMM.Get_size()
        rank = self.COMM.Get_rank()
        for root in range(size):
            for op in (
                MPI.SUM,
                MPI.PROD,
                MPI.MAX,
                MPI.MIN,
                MPI.MAXLOC,
                MPI.MINLOC,
                MPI.REPLACE,
                MPI.NO_OP,
            ):
                if op == MPI.OP_NULL:
                    continue
                if op in (MPI.MAXLOC, MPI.MINLOC):
                    sendobj = (rank, rank)
                else:
                    sendobj = rank
                value = self.COMM.reduce(sendobj, op=op, root=root)
                if rank != root:
                    self.assertIsNone(value)
                else:
                    if op == MPI.SUM:
                        self.assertEqual(value, cumsum(range(size)))
                    elif op == MPI.PROD:
                        self.assertEqual(value, cumprod(range(size)))
                    elif op == MPI.MAX:
                        self.assertEqual(value, size - 1)
                    elif op == MPI.MIN:
                        self.assertEqual(value, 0)
                    elif op == MPI.MAXLOC:
                        self.assertEqual(value[0], size - 1)
                        self.assertEqual(value[1], size - 1)
                    elif op == MPI.MINLOC:
                        self.assertEqual(value[0], 0)
                        self.assertEqual(value[1], 0)
                    elif op == MPI.REPLACE:
                        self.assertEqual(value, size - 1)
                    elif op == MPI.NO_OP:
                        self.assertEqual(value, 0)
        badroots = {-size, size}
        for root in badroots:
            with self.assertRaises(MPI.Exception) as cm:
                self.COMM.reduce(None, op=MPI.NO_OP, root=root)
            ierr = cm.exception.Get_error_class()
            self.assertEqual(ierr, MPI.ERR_ROOT)

    def testAllreduce(self):
        size = self.COMM.Get_size()
        rank = self.COMM.Get_rank()
        for op in (
            MPI.SUM,
            MPI.PROD,
            MPI.MAX,
            MPI.MIN,
            MPI.MAXLOC,
            MPI.MINLOC,
            MPI.REPLACE,
            MPI.NO_OP,
        ):
            if op == MPI.OP_NULL:
                continue
            if op in (MPI.MAXLOC, MPI.MINLOC):
                sendobj = (rank, rank)
            else:
                sendobj = rank
            value = self.COMM.allreduce(sendobj, op)
            if op == MPI.SUM:
                self.assertEqual(value, cumsum(range(size)))
            elif op == MPI.PROD:
                self.assertEqual(value, cumprod(range(size)))
            elif op == MPI.MAX:
                self.assertEqual(value, size - 1)
            elif op == MPI.MIN:
                self.assertEqual(value, 0)
            elif op == MPI.MAXLOC:
                self.assertEqual(value[1], size - 1)
            elif op == MPI.MINLOC:
                self.assertEqual(value[1], 0)
            elif op == MPI.REPLACE:
                self.assertEqual(value, size - 1)
            elif op == MPI.NO_OP:
                self.assertEqual(value, 0)

    def testScan(self):
        size = self.COMM.Get_size()
        rank = self.COMM.Get_rank()
        # --
        sscan = self.COMM.scan(size, op=MPI.SUM)
        self.assertEqual(sscan, cumsum([size] * (rank + 1)))
        # --
        rscan = self.COMM.scan(rank, op=MPI.SUM)
        self.assertEqual(rscan, cumsum(range(rank + 1)))
        # --
        minloc = self.COMM.scan((rank, rank), op=MPI.MINLOC)
        maxloc = self.COMM.scan((rank, rank), op=MPI.MAXLOC)
        self.assertEqual(minloc, (0, 0))
        self.assertEqual(maxloc, (rank, rank))
        # --
        if MPI.REPLACE != MPI.OP_NULL:
            rscan = self.COMM.scan(rank, op=MPI.REPLACE)
            self.assertEqual(rscan, rank)
        # --
        if MPI.NO_OP != MPI.OP_NULL:
            rscan = self.COMM.scan(rank, op=MPI.NO_OP)
            self.assertEqual(rscan, 0)

    def testExscan(self):
        size = self.COMM.Get_size()
        rank = self.COMM.Get_rank()
        # --
        sscan = self.COMM.exscan(size, op=MPI.SUM)
        if rank == 0:
            self.assertIsNone(sscan)
        else:
            self.assertEqual(sscan, cumsum([size] * (rank)))
        # --
        rscan = self.COMM.exscan(rank, op=MPI.SUM)
        if rank == 0:
            self.assertIsNone(sscan)
        else:
            self.assertEqual(rscan, cumsum(range(rank)))
        # --
        minloc = self.COMM.exscan((rank, rank), op=MPI.MINLOC)
        maxloc = self.COMM.exscan((rank, rank), op=MPI.MAXLOC)
        if rank == 0:
            self.assertIsNone(minloc)
            self.assertIsNone(maxloc)
        else:
            self.assertEqual(minloc, (0, 0))
            self.assertEqual(maxloc, (rank - 1, rank - 1))
        # --
        if MPI.REPLACE != MPI.OP_NULL:
            rscan = self.COMM.exscan(rank, op=MPI.REPLACE)
            if rank == 0:
                self.assertIsNone(rscan)
            else:
                self.assertEqual(rscan, rank - 1)
        # --
        if MPI.NO_OP != MPI.OP_NULL:
            rscan = self.COMM.exscan(rank, op=MPI.NO_OP)
            if rank == 0:
                self.assertIsNone(rscan)
            else:
                self.assertEqual(rscan, 0)


class TestCCOObjSelf(BaseTestCCOObj, unittest.TestCase):
    #
    COMM = MPI.COMM_SELF


class TestCCOObjWorld(BaseTestCCOObj, unittest.TestCase):
    #
    COMM = MPI.COMM_WORLD


class TestCCOObjSelfDup(TestCCOObjSelf):
    #
    def setUp(self):
        self.COMM = MPI.COMM_SELF.Dup()

    def tearDown(self):
        self.COMM.Free()


@unittest.skipMPI("openmpi(<1.4.0)", MPI.Query_thread() > MPI.THREAD_SINGLE)
class TestCCOObjWorldDup(TestCCOObjWorld):
    #
    def setUp(self):
        self.COMM = MPI.COMM_WORLD.Dup()

    def tearDown(self):
        self.COMM.Free()


if __name__ == "__main__":
    unittest.main()



# Storage order for arrays
# ------------------------
ORDER_C       = MPI_ORDER_C        #: C order (a.k.a. row major)
ORDER_FORTRAN = MPI_ORDER_FORTRAN  #: Fortran order (a.k.a. column major)
ORDER_F       = MPI_ORDER_FORTRAN  #: Convenience alias for ORDER_FORTRAN


# Type classes for Fortran datatype matching
# ------------------------------------------
TYPECLASS_LOGICAL = MPI_TYPECLASS_LOGICAL
TYPECLASS_INTEGER = MPI_TYPECLASS_INTEGER
TYPECLASS_REAL    = MPI_TYPECLASS_REAL
TYPECLASS_COMPLEX = MPI_TYPECLASS_COMPLEX


# Type of distributions (HPF-like arrays)
# ---------------------------------------
DISTRIBUTE_NONE      = MPI_DISTRIBUTE_NONE       #: Dimension not distributed
DISTRIBUTE_BLOCK     = MPI_DISTRIBUTE_BLOCK      #: Block distribution
DISTRIBUTE_CYCLIC    = MPI_DISTRIBUTE_CYCLIC     #: Cyclic distribution
DISTRIBUTE_DFLT_DARG = MPI_DISTRIBUTE_DFLT_DARG  #: Default distribution


# Combiner values for datatype decoding
# -------------------------------------
COMBINER_NAMED            = MPI_COMBINER_NAMED
COMBINER_DUP              = MPI_COMBINER_DUP
COMBINER_CONTIGUOUS       = MPI_COMBINER_CONTIGUOUS
COMBINER_VECTOR           = MPI_COMBINER_VECTOR
COMBINER_HVECTOR          = MPI_COMBINER_HVECTOR
COMBINER_INDEXED          = MPI_COMBINER_INDEXED
COMBINER_HINDEXED         = MPI_COMBINER_HINDEXED
COMBINER_INDEXED_BLOCK    = MPI_COMBINER_INDEXED_BLOCK
COMBINER_HINDEXED_BLOCK   = MPI_COMBINER_HINDEXED_BLOCK
COMBINER_STRUCT           = MPI_COMBINER_STRUCT
COMBINER_SUBARRAY         = MPI_COMBINER_SUBARRAY
COMBINER_DARRAY           = MPI_COMBINER_DARRAY
COMBINER_RESIZED          = MPI_COMBINER_RESIZED
COMBINER_VALUE_INDEX      = MPI_COMBINER_VALUE_INDEX
COMBINER_F90_INTEGER      = MPI_COMBINER_F90_INTEGER
COMBINER_F90_REAL         = MPI_COMBINER_F90_REAL
COMBINER_F90_COMPLEX      = MPI_COMBINER_F90_COMPLEX


cdef class Datatype:
    """
    Datatype object.
    """

    def __cinit__(self, Datatype datatype: Datatype | None = None):
        cinit(self, datatype)

    def __dealloc__(self):
        dealloc(self)

    def __richcmp__(self, other, int op):
        if not isinstance(other, Datatype): return NotImplemented
        return richcmp(self, other, op)

    def __bool__(self) -> bool:
        return nonnull(self)

    def __reduce__(self) -> str | tuple[Any, ...]:
        return reduce_Datatype(self)

    property handle:
        """MPI handle."""
        def __get__(self) -> int:
            return tohandle(self)

    @classmethod
    def fromhandle(cls, handle: int) -> Datatype:
        """
        Create object from MPI handle.
        """
        return fromhandle(<MPI_Datatype> <Py_uintptr_t> handle)

    def free(self) -> None:
        """
        Call `Free` if not null or predefined.
        """
        safefree(self)

    # Datatype Accessors
    # ------------------

    def Get_size(self) -> int:
        """
        Return the number of bytes occupied by entries in the datatype.
        """
        cdef MPI_Count size = 0
        CHKERR( MPI_Type_size_c(self.ob_mpi, &size) )
        return size

    property size:
        """Size (in bytes)."""
        def __get__(self) -> int:
            cdef MPI_Count size = 0
            CHKERR( MPI_Type_size_c(self.ob_mpi, &size) )
            return size

    def Get_extent(self) -> tuple[int, int]:
        """
        Return lower bound and extent of datatype.
        """
        cdef MPI_Count lb = 0, extent = 0
        CHKERR( MPI_Type_get_extent_c(self.ob_mpi, &lb, &extent) )
        return (lb, extent)

    property extent:
        """Extent."""
        def __get__(self) -> int:
            cdef MPI_Count lb = 0, extent = 0
            CHKERR( MPI_Type_get_extent_c(self.ob_mpi, &lb, &extent) )
            return extent

    property lb:
        """Lower bound."""
        def __get__(self) -> int:
            cdef MPI_Count lb = 0, extent = 0
            CHKERR( MPI_Type_get_extent_c(self.ob_mpi, &lb, &extent) )
            return lb

    property ub:
        """Upper bound."""
        def __get__(self) -> int:
            cdef MPI_Count lb = 0, extent = 0
            CHKERR( MPI_Type_get_extent_c(self.ob_mpi, &lb, &extent) )
            return lb + extent

    # Datatype Constructors
    # ---------------------

    def Dup(self) -> Self:
        """
        Duplicate a datatype.
        """
        cdef Datatype datatype = <Datatype>New(type(self))
        CHKERR( MPI_Type_dup(self.ob_mpi, &datatype.ob_mpi) )
        return datatype

    Create_dup = Dup  #: convenience alias

    def Create_contiguous(self, Count count: int) -> Self:
        """
        Create a contiguous datatype.
        """
        cdef Datatype datatype = <Datatype>New(type(self))
        CHKERR( MPI_Type_contiguous_c(
            count,
            self.ob_mpi, &datatype.ob_mpi) )
        return datatype

    def Create_vector(
        self,
        Count count: int,
        Count blocklength: int,
        Count stride: int,
    ) -> Self:
        """
        Create a vector (strided) datatype.
        """
        cdef Datatype datatype = <Datatype>New(type(self))
        CHKERR( MPI_Type_vector_c(
            count, blocklength, stride,
            self.ob_mpi, &datatype.ob_mpi) )
        return datatype

    def Create_hvector(
        self,
        Count count: int,
        Count blocklength: int,
        Count stride: int,
    ) -> Self:
        """
        Create a vector (strided) datatype with stride in bytes.
        """
        cdef Datatype datatype = <Datatype>New(type(self))
        CHKERR( MPI_Type_create_hvector_c(
            count, blocklength, stride,
            self.ob_mpi, &datatype.ob_mpi) )
        return datatype

    def Create_indexed(
        self,
        blocklengths: Sequence[int],
        displacements: Sequence[int],
    ) -> Self:
        """
        Create an indexed datatype.
        """
        cdef MPI_Count count = 0, *iblen = NULL, *idisp = NULL
        blocklengths  = getarray(blocklengths,  &count, &iblen)
        displacements = chkarray(displacements,  count, &idisp)
        #
        cdef Datatype datatype = <Datatype>New(type(self))
        CHKERR( MPI_Type_indexed_c(
            count, iblen, idisp,
            self.ob_mpi, &datatype.ob_mpi) )
        return datatype

    def Create_hindexed(
        self,
        blocklengths: Sequence[int],
        displacements: Sequence[int],
    ) -> Self:
        """
        Create an indexed datatype.

        .. note:: Displacements are measured in bytes.
        """
        cdef MPI_Count count = 0, *iblen = NULL, *idisp = NULL
        blocklengths = getarray(blocklengths, &count, &iblen)
        displacements = chkarray(displacements, count, &idisp)
        #
        cdef Datatype datatype = <Datatype>New(type(self))
        CHKERR( MPI_Type_create_hindexed_c(
            count, iblen, idisp,
            self.ob_mpi,
            &datatype.ob_mpi) )
        return datatype

    def Create_indexed_block(
        self,
        Count blocklength: int,
        displacements: Sequence[int],
    ) -> Self:
        """
        Create an indexed datatype with constant-sized blocks.
        """
        cdef Count count = 0, *idisp = NULL
        displacements = getarray(displacements, &count, &idisp)
        #
        cdef Datatype datatype = <Datatype>New(type(self))
        CHKERR( MPI_Type_create_indexed_block_c(
            count, blocklength, idisp,
            self.ob_mpi, &datatype.ob_mpi) )
        return datatype

    def Create_hindexed_block(
        self,
        Count blocklength: int,
        displacements: Sequence[int],
    ) -> Self:
        """
        Create an indexed datatype with constant-sized blocks.

        .. note:: Displacements are measured in bytes.
        """
        cdef MPI_Count count = 0, *idisp = NULL
        displacements = getarray(displacements, &count, &idisp)
        #
        cdef Datatype datatype = <Datatype>New(type(self))
        CHKERR( MPI_Type_create_hindexed_block_c(
            count, blocklength, idisp,
            self.ob_mpi, &datatype.ob_mpi) )
        return datatype

    @classmethod
    def Create_struct(
        cls,
        blocklengths: Sequence[int],
        displacements: Sequence[int],
        datatypes: Sequence[Datatype],
    ) -> Self:
        """
        Create a general composite (struct) datatype.

        .. note:: Displacements are measured in bytes.
        """
        cdef MPI_Count count = 0, *iblen = NULL, *idisp = NULL
        cdef MPI_Datatype *ptype = NULL
        blocklengths = getarray(blocklengths, &count, &iblen)
        displacements = chkarray(displacements, count, &idisp)
        datatypes = asarray_Datatype(datatypes, count, &ptype)
        #
        cdef Datatype datatype = <Datatype>New(cls)
        CHKERR( MPI_Type_create_struct_c(
            count, iblen, idisp, ptype,
            &datatype.ob_mpi) )
        return datatype

    # Subarray Datatype Constructor
    # -----------------------------

    def Create_subarray(
        self,
        sizes: Sequence[int],
        subsizes: Sequence[int],
        starts: Sequence[int],
        int order: int = ORDER_C,
    ) -> Self:
        """
        Create a datatype for a subarray of a multidimensional array.
        """
        cdef int ndims = 0
        cdef MPI_Count *isizes = NULL
        cdef MPI_Count *isubsizes = NULL
        cdef MPI_Count *istarts = NULL
        sizes    = getarray(sizes,   &ndims, &isizes   )
        subsizes = chkarray(subsizes, ndims, &isubsizes)
        starts   = chkarray(starts,   ndims, &istarts  )
        #
        cdef Datatype datatype = <Datatype>New(type(self))
        CHKERR( MPI_Type_create_subarray_c(
            ndims, isizes, isubsizes, istarts, order,
            self.ob_mpi, &datatype.ob_mpi) )
        return datatype

    # Distributed Array Datatype Constructor
    # --------------------------------------

    def Create_darray(
        self,
        int size: int,
        int rank: int,
        gsizes: Sequence[int],
        distribs: Sequence[int],
        dargs: Sequence[int],
        psizes: Sequence[int],
        int order: int = ORDER_C,
    ) -> Self:
        """
        Create a datatype for a distributed array on Cartesian process grids.
        """
        cdef int ndims = 0
        cdef MPI_Count *igsizes = NULL
        cdef int *idistribs = NULL, *idargs = NULL, *ipsizes = NULL
        gsizes   = getarray(gsizes,  &ndims, &igsizes   )
        distribs = chkarray(distribs, ndims, &idistribs )
        dargs    = chkarray(dargs,    ndims, &idargs    )
        psizes   = chkarray(psizes,   ndims, &ipsizes   )
        #
        cdef Datatype datatype = <Datatype>New(type(self))
        CHKERR( MPI_Type_create_darray_c(
            size, rank, ndims, igsizes,
            idistribs, idargs, ipsizes, order,
            self.ob_mpi, &datatype.ob_mpi) )
        return datatype

    # Pair Datatype
    # -------------

    @classmethod
    def Get_value_index(
        cls,
        Datatype value: Datatype,
        Datatype index: Datatype,
    ) -> Self:
        """
        Return a predefined pair datatype.
        """
        cdef Datatype datatype = <Datatype>New(cls)
        CHKERR( MPI_Type_get_value_index(
            value.ob_mpi, index.ob_mpi, &datatype.ob_mpi) )
        return datatype

    # Parametrized and size-specific Fortran Datatypes
    # ------------------------------------------------

    @classmethod
    def Create_f90_integer(cls, int r: int) -> Self:
        """
        Return a bounded integer datatype.
        """
        cdef Datatype datatype = <Datatype>New(cls)
        CHKERR( MPI_Type_create_f90_integer(r, &datatype.ob_mpi) )
        return datatype

    @classmethod
    def Create_f90_real(cls, int p: int, int r: int) -> Self:
        """
        Return a bounded real datatype.
        """
        cdef Datatype datatype = <Datatype>New(cls)
        CHKERR( MPI_Type_create_f90_real(p, r, &datatype.ob_mpi) )
        return datatype

    @classmethod
    def Create_f90_complex(cls, int p: int, int r: int) -> Self:
        """
        Return a bounded complex datatype.
        """
        cdef Datatype datatype = <Datatype>New(cls)
        CHKERR( MPI_Type_create_f90_complex(p, r, &datatype.ob_mpi) )
        return datatype

    @classmethod
    def Match_size(cls, int typeclass: int, int size: int) -> Self:
        """
        Find a datatype matching a specified size in bytes.
        """
        cdef Datatype datatype = <Datatype>New(cls)
        CHKERR( MPI_Type_match_size(typeclass, size, &datatype.ob_mpi) )
        return datatype

    # Use of Derived Datatypes
    # ------------------------

    def Commit(self) -> Self:
        """
        Commit the datatype.
        """
        CHKERR( MPI_Type_commit(&self.ob_mpi) )
        return self

    def Free(self) -> None:
        """
        Free the datatype.
        """
        cdef MPI_Datatype save = self.ob_mpi
        CHKERR( MPI_Type_free(&self.ob_mpi) )
        if constobj(self): self.ob_mpi = save

    # Datatype Resizing
    # -----------------

    def Create_resized(self, Count lb: int, Count extent: int) -> Self:
        """
        Create a datatype with a new lower bound and extent.
        """
        cdef Datatype datatype = <Datatype>New(type(self))
        CHKERR( MPI_Type_create_resized_c(
            self.ob_mpi, lb, extent, &datatype.ob_mpi) )
        return datatype

    Resized = Create_resized  #: compatibility alias

    def Get_true_extent(self) -> tuple[int, int]:
        """
        Return the true lower bound and extent of a datatype.
        """
        cdef MPI_Count lb = 0, extent = 0
        CHKERR( MPI_Type_get_true_extent_c(
            self.ob_mpi, &lb, &extent) )
        return (lb, extent)

    property true_extent:
        """True extent."""
        def __get__(self) -> int:
            cdef MPI_Count lb = 0, extent = 0
            CHKERR( MPI_Type_get_true_extent_c(
                self.ob_mpi, &lb, &extent) )
            return extent

    property true_lb:
        """True lower bound."""
        def __get__(self) -> int:
            cdef MPI_Count lb = 0, extent = 0
            CHKERR( MPI_Type_get_true_extent_c(
                self.ob_mpi, &lb, &extent) )
            return lb

    property true_ub:
        """True upper bound."""
        def __get__(self) -> int:
            cdef MPI_Count lb = 0, extent = 0
            CHKERR( MPI_Type_get_true_extent_c(
                self.ob_mpi, &lb, &extent) )
            return lb + extent

    # Decoding a Datatype
    # -------------------

    def Get_envelope(self) -> tuple[int, int, int, int, int]:
        """
        Return the number of input arguments used to create a datatype.
        """
        cdef int combiner = MPI_UNDEFINED
        cdef MPI_Count ni = 0, na = 0, nc = 0, nd = 0
        CHKERR( MPI_Type_get_envelope_c(
            self.ob_mpi, &ni, &na, &nc, &nd, &combiner) )
        return (ni, na, nc, nd, combiner)

    property envelope:
        """Envelope."""
        def __get__(self) -> tuple[int, int, int, int, int]:
            return self.Get_envelope()

    def Get_contents(self) -> \
            tuple[list[int], list[int], list[int], list[Datatype]]:
        """
        Return the input arguments used to create a datatype.
        """
        cdef int combiner = MPI_UNDEFINED
        cdef MPI_Count ni = 0, na = 0, nc = 0, nd = 0
        CHKERR( MPI_Type_get_envelope_c(
            self.ob_mpi, &ni, &na, &nc, &nd, &combiner) )
        cdef int *i = NULL
        cdef MPI_Aint *a = NULL
        cdef MPI_Count *c = NULL
        cdef MPI_Datatype *d = NULL
        cdef unused1 = allocate(ni, sizeof(int), &i)
        cdef unused2 = allocate(na, sizeof(MPI_Aint), &a)
        cdef unused3 = allocate(nc, sizeof(MPI_Count), &c)
        cdef unused4 = allocate(nd, sizeof(MPI_Datatype), &d)
        CHKERR( MPI_Type_get_contents_c(
            self.ob_mpi, ni, na, nc, nd, i, a, c, d) )
        cdef object integers  = [i[k] for k in range(ni)]
        cdef object addresses = [a[k] for k in range(na)]
        cdef object counts    = [c[k] for k in range(nc)]
        cdef object datatypes = [ref_Datatype(d[k]) for k in range(nd)]
        return (integers, addresses, counts, datatypes)

    property contents:
        """Contents."""
        def __get__(self) -> \
                tuple[list[int], list[int], list[int], list[Datatype]]:
            return self.Get_contents()

    def decode(self) -> tuple[Datatype, str, dict[str, Any]]:
        """
        Convenience method for decoding a datatype.
        """
        return datatype_decode(self, mark=False)

    property combiner:
        """Combiner."""
        def __get__(self) -> int:
            cdef int combiner = self.Get_envelope()[-1]
            return combiner

    property is_named:
        """Is a named datatype."""
        def __get__(self) -> bool:
            cdef int combiner = self.Get_envelope()[-1]
            return combiner == MPI_COMBINER_NAMED

    property is_predefined:
        """Is a predefined datatype."""
        def __get__(self) -> bool:
            if self.ob_mpi == MPI_DATATYPE_NULL: return True
            cdef int combiner = self.Get_envelope()[-1]
            return (combiner == MPI_COMBINER_NAMED or
                    combiner == MPI_COMBINER_VALUE_INDEX or
                    combiner == MPI_COMBINER_F90_INTEGER or
                    combiner == MPI_COMBINER_F90_REAL or
                    combiner == MPI_COMBINER_F90_COMPLEX)

    # Pack and Unpack
    # ---------------

    def Pack(
        self,
        inbuf: BufSpec,
        outbuf: BufSpec,
        Count position: int,
        Comm comm: Comm,
    ) -> int:
        """
        Pack into contiguous memory according to datatype.
        """
        cdef MPI_Count lb = 0, extent = 0
        CHKERR( MPI_Type_get_extent_c(self.ob_mpi, &lb, &extent) )
        #
        cdef void *ibptr = NULL, *obptr = NULL
        cdef MPI_Aint iblen = 0, oblen = 0
        cdef unused1 = asbuffer_r(inbuf,  &ibptr, &iblen)
        cdef unused2 = asbuffer_w(outbuf, &obptr, &oblen)
        cdef MPI_Count icount = iblen // extent
        cdef MPI_Count ocount = oblen
        #
        CHKERR( MPI_Pack_c(
            ibptr, icount, self.ob_mpi,
            obptr, ocount, &position,
            comm.ob_mpi) )
        return position

    def Unpack(
        self,
        inbuf: BufSpec,
        Count position: int,
        outbuf: BufSpec,
        Comm comm: Comm,
    ) -> int:
        """
        Unpack from contiguous memory according to datatype.
        """
        cdef MPI_Count lb = 0, extent = 0
        CHKERR( MPI_Type_get_extent_c(self.ob_mpi, &lb, &extent) )
        #
        cdef void *ibptr = NULL, *obptr = NULL
        cdef MPI_Aint iblen = 0, oblen = 0
        cdef unused1 = asbuffer_r(inbuf,  &ibptr, &iblen)
        cdef unused2 = asbuffer_w(outbuf, &obptr, &oblen)
        cdef MPI_Count icount = iblen
        cdef MPI_Count ocount = oblen // extent
        #
        CHKERR( MPI_Unpack_c(
            ibptr, icount, &position,
            obptr, ocount, self.ob_mpi,
            comm.ob_mpi) )
        return position

    def Pack_size(
        self,
        Count count: int,
        Comm comm: Comm,
    ) -> int:
        """
        Determine the amount of space needed to pack a message.

        .. note:: Returns an upper bound measured in bytes.
        """
        cdef MPI_Count size = 0
        CHKERR( MPI_Pack_size_c(
            count, self.ob_mpi,
            comm.ob_mpi, &size) )
        return size

    # Canonical Pack and Unpack
    # -------------------------

    def Pack_external(
        self,
        datarep: str,
        inbuf: BufSpec,
        outbuf: BufSpec,
        Count position: int,
    ) -> int:
        """
        Pack into contiguous memory according to datatype.

        Uses the portable data representation **external32**.
        """
        cdef char *cdatarep = NULL
        datarep = asmpistr(datarep, &cdatarep)
        cdef MPI_Count lb = 0, extent = 0
        CHKERR( MPI_Type_get_extent_c(self.ob_mpi, &lb, &extent) )
        #
        cdef void *ibptr = NULL, *obptr = NULL
        cdef MPI_Aint iblen = 0, oblen = 0
        cdef unused1 = asbuffer_r(inbuf,  &ibptr, &iblen)
        cdef unused2 = asbuffer_w(outbuf, &obptr, &oblen)
        cdef MPI_Count icount = iblen // extent
        cdef MPI_Count ocount = oblen
        #
        CHKERR( MPI_Pack_external_c(
            cdatarep,
            ibptr, icount, self.ob_mpi,
            obptr, ocount, &position) )
        return position

    def Unpack_external(
        self,
        datarep: str,
        inbuf: BufSpec,
        Count position: int,
        outbuf: BufSpec,
    ) -> int:
        """
        Unpack from contiguous memory according to datatype.

        Uses the portable data representation **external32**.
        """
        cdef char *cdatarep = NULL
        datarep = asmpistr(datarep, &cdatarep)
        cdef MPI_Count lb = 0, extent = 0
        CHKERR( MPI_Type_get_extent_c(self.ob_mpi, &lb, &extent) )
        #
        cdef void *ibptr = NULL, *obptr = NULL
        cdef MPI_Aint iblen = 0, oblen = 0
        cdef unused1 = asbuffer_r(inbuf,  &ibptr, &iblen)
        cdef unused2 = asbuffer_w(outbuf, &obptr, &oblen)
        cdef MPI_Count icount = iblen
        cdef MPI_Count ocount = oblen // extent
        #
        CHKERR( MPI_Unpack_external_c(
            cdatarep,
            ibptr, icount, &position,
            obptr, ocount, self.ob_mpi) )
        return position

    def Pack_external_size(
        self,
        datarep: str,
        Count count: int,
    ) -> int:
        """
        Determine the amount of space needed to pack a message.

        Uses the portable data representation **external32**.

        .. note:: Returns an upper bound measured in bytes.
        """
        cdef char *cdatarep = NULL
        cdef MPI_Count size = 0
        datarep = asmpistr(datarep, &cdatarep)
        CHKERR( MPI_Pack_external_size_c(
            cdatarep, count, self.ob_mpi, &size) )
        return size

    # Attributes
    # ----------

    def Get_attr(self, int keyval: int) -> int | Any | None:
        """
        Retrieve attribute value by key.
        """
        cdef void *attrval = NULL
        cdef int flag = 0
        CHKERR( MPI_Type_get_attr(self.ob_mpi, keyval, &attrval, &flag) )
        if not flag: return None
        if attrval == NULL: return 0
        # user-defined attribute keyval
        return PyMPI_attr_get(self.ob_mpi, keyval, attrval)

    def Set_attr(self, int keyval: int, attrval: Any) -> None:
        """
        Store attribute value associated with a key.
        """
        PyMPI_attr_set(self.ob_mpi, keyval, attrval)

    def Delete_attr(self, int keyval: int) -> None:
        """
        Delete attribute value associated with a key.
        """
        PyMPI_attr_del(self.ob_mpi, keyval)

    @classmethod
    def Create_keyval(
        cls,
        copy_fn: Callable[[Datatype, int, Any], Any] | None = None,
        delete_fn: Callable[[Datatype, int, Any], None] | None = None,
        nopython: bool = False,
    ) -> int:
        """
        Create a new attribute key for datatypes.
        """
        cdef int keyval = MPI_KEYVAL_INVALID
        cdef MPI_Type_copy_attr_function *_copy = PyMPI_attr_copy_fn
        cdef MPI_Type_delete_attr_function *_del = PyMPI_attr_delete_fn
        cdef _p_keyval state = _p_keyval(copy_fn, delete_fn, nopython)
        CHKERR( MPI_Type_create_keyval(_copy, _del, &keyval, <void *>state) )
        PyMPI_attr_state_set(MPI_DATATYPE_NULL, keyval, state)
        return keyval

    @classmethod
    def Free_keyval(cls, int keyval: int) -> int:
        """
        Free an attribute key for datatypes.
        """
        cdef int keyval_save = keyval
        CHKERR( MPI_Type_free_keyval(&keyval) )
        PyMPI_attr_state_del(MPI_DATATYPE_NULL, keyval_save)
        return keyval

    # Naming Objects
    # --------------

    def Get_name(self) -> str:
        """
        Get the print name for this datatype.
        """
        cdef char name[MPI_MAX_OBJECT_NAME+1]
        cdef int nlen = 0
        CHKERR( MPI_Type_get_name(self.ob_mpi, name, &nlen) )
        return tompistr(name, nlen)

    def Set_name(self, name: str) -> None:
        """
        Set the print name for this datatype.
        """
        cdef char *cname = NULL
        name = asmpistr(name, &cname)
        CHKERR( MPI_Type_set_name(self.ob_mpi, cname) )

    property name:
        """Print name."""
        def __get__(self) -> str:
            return self.Get_name()

        def __set__(self, value: str):
            self.Set_name(value)

    # Integer Handle
    # --------------

    def toint(self) -> int:
        """
        Translate object to integer handle.
        """
        return MPI_Type_toint(self.ob_mpi)

    @classmethod
    def fromint(cls, arg: int, /) -> Datatype:
        """
        Translate integer handle to object.
        """
        return fromhandle(MPI_Type_fromint(arg))

    # Fortran Handle
    # --------------

    def py2f(self) -> int:
        """
        """
        return MPI_Type_c2f(self.ob_mpi)

    @classmethod
    def f2py(cls, arg: int) -> Datatype:
        """
        """
        return fromhandle(MPI_Type_f2c(arg))

    # Python/NumPy interoperability
    # -----------------------------

    def tocode(self) -> str:
        """
        Get character code or type string from predefined MPI datatype.
        """
        cdef const char *s = DatatypeCode(self.ob_mpi)
        if s != NULL: return pystr(s)
        raise ValueError("cannot map to character code or type string")

    @classmethod
    def fromcode(cls, code: str) -> Datatype:
        """
        Get predefined MPI datatype from character code or type string.
        """
        try:
            return <Datatype?> TypeDict[code]
        except KeyError:
            raise ValueError(f"cannot map code {code!r} to MPI datatype")

    property typestr:
        """Type string."""
        def __get__(self) -> str:
            if self.ob_mpi == MPI_DATATYPE_NULL: return ""
            cdef const char *s = DatatypeStr(self.ob_mpi)
            if s != NULL: return pystr(s)
            return f"V{mpiextent(self.ob_mpi)}"

    property typechar:
        """Character code."""
        def __get__(self) -> str:
            if self.ob_mpi == MPI_DATATYPE_NULL: return ""
            cdef const char *s = DatatypeChar(self.ob_mpi)
            if s != NULL: return pystr(s)
            return "V"


# Address Functions
# -----------------

def Get_address(location: Buffer | Bottom) -> int:
    """
    Get the address of a location in memory.
    """
    cdef void *baseptr = MPI_BOTTOM
    if not is_BOTTOM(location):
        asbuffer_r(location, &baseptr, NULL)
    cdef MPI_Aint address = 0
    CHKERR( MPI_Get_address(baseptr, &address) )
    return address


def Aint_add(Aint base: int, Aint disp: int) -> int:
    """
    Return the sum of base address and displacement.
    """
    return MPI_Aint_add(base, disp)


def Aint_diff(Aint addr1: int, Aint addr2: int) -> int:
    """
    Return the difference between absolute addresses.
    """
    return MPI_Aint_diff(addr1, addr2)


cdef Datatype __DATATYPE_NULL__ = def_Datatype( MPI_DATATYPE_NULL , "DATATYPE_NULL" )  # noqa

cdef Datatype __PACKED__ = def_Datatype( MPI_PACKED , "PACKED" )
cdef Datatype __BYTE__   = def_Datatype( MPI_BYTE   , "BYTE"   )
cdef Datatype __AINT__   = def_Datatype( MPI_AINT   , "AINT"   )
cdef Datatype __OFFSET__ = def_Datatype( MPI_OFFSET , "OFFSET" )
cdef Datatype __COUNT__  = def_Datatype( MPI_COUNT  , "COUNT"  )

cdef Datatype __CHAR__               = def_Datatype( MPI_CHAR               , "CHAR"               )  # noqa
cdef Datatype __WCHAR__              = def_Datatype( MPI_WCHAR              , "WCHAR"              )  # noqa
cdef Datatype __SIGNED_CHAR__        = def_Datatype( MPI_SIGNED_CHAR        , "SIGNED_CHAR"        )  # noqa
cdef Datatype __SHORT__              = def_Datatype( MPI_SHORT              , "SHORT"              )  # noqa
cdef Datatype __INT__                = def_Datatype( MPI_INT                , "INT"                )  # noqa
cdef Datatype __LONG__               = def_Datatype( MPI_LONG               , "LONG"               )  # noqa
cdef Datatype __LONG_LONG__          = def_Datatype( MPI_LONG_LONG          , "LONG_LONG"          )  # noqa
cdef Datatype __UNSIGNED_CHAR__      = def_Datatype( MPI_UNSIGNED_CHAR      , "UNSIGNED_CHAR"      )  # noqa
cdef Datatype __UNSIGNED_SHORT__     = def_Datatype( MPI_UNSIGNED_SHORT     , "UNSIGNED_SHORT"     )  # noqa
cdef Datatype __UNSIGNED__           = def_Datatype( MPI_UNSIGNED           , "UNSIGNED"           )  # noqa
cdef Datatype __UNSIGNED_LONG__      = def_Datatype( MPI_UNSIGNED_LONG      , "UNSIGNED_LONG"      )  # noqa
cdef Datatype __UNSIGNED_LONG_LONG__ = def_Datatype( MPI_UNSIGNED_LONG_LONG , "UNSIGNED_LONG_LONG" )  # noqa
cdef Datatype __FLOAT__              = def_Datatype( MPI_FLOAT              , "FLOAT"              )  # noqa
cdef Datatype __DOUBLE__             = def_Datatype( MPI_DOUBLE             , "DOUBLE"             )  # noqa
cdef Datatype __LONG_DOUBLE__        = def_Datatype( MPI_LONG_DOUBLE        , "LONG_DOUBLE"        )  # noqa

cdef Datatype __C_BOOL__                = def_Datatype( MPI_C_BOOL   , "C_BOOL"   )  # noqa
cdef Datatype __INT8_T__                = def_Datatype( MPI_INT8_T   , "INT8_T"   )  # noqa
cdef Datatype __INT16_T__               = def_Datatype( MPI_INT16_T  , "INT16_T"  )  # noqa
cdef Datatype __INT32_T__               = def_Datatype( MPI_INT32_T  , "INT32_T"  )  # noqa
cdef Datatype __INT64_T__               = def_Datatype( MPI_INT64_T  , "INT64_T"  )  # noqa
cdef Datatype __UINT8_T__               = def_Datatype( MPI_UINT8_T  , "UINT8_T"  )  # noqa
cdef Datatype __UINT16_T__              = def_Datatype( MPI_UINT16_T , "UINT16_T" )  # noqa
cdef Datatype __UINT32_T__              = def_Datatype( MPI_UINT32_T , "UINT32_T" )  # noqa
cdef Datatype __UINT64_T__              = def_Datatype( MPI_UINT64_T , "UINT64_T" )  # noqa

cdef Datatype __FLOAT16_T__             = def_Datatype( MPI_FLOAT16_T             , "FLOAT16_T"             )  # noqa
cdef Datatype __FLOAT32_T__             = __FLOAT__
cdef Datatype __FLOAT64_T__             = __DOUBLE__
cdef Datatype __BFLOAT16_T__            = def_Datatype( MPI_BFLOAT16_T            , "BFLOAT16_T"            )  # noqa
cdef Datatype __C_COMPLEX__             = def_Datatype( MPI_C_COMPLEX             , "C_COMPLEX"             )  # noqa
cdef Datatype __C_FLOAT_COMPLEX__       = __C_COMPLEX__
cdef Datatype __C_DOUBLE_COMPLEX__      = def_Datatype( MPI_C_DOUBLE_COMPLEX      , "C_DOUBLE_COMPLEX"      )  # noqa
cdef Datatype __C_LONG_DOUBLE_COMPLEX__ = def_Datatype( MPI_C_LONG_DOUBLE_COMPLEX , "C_LONG_DOUBLE_COMPLEX" )  # noqa

cdef Datatype __CXX_BOOL__                = def_Datatype( MPI_CXX_BOOL                , "CXX_BOOL"                )  # noqa
cdef Datatype __CXX_FLOAT_COMPLEX__       = def_Datatype( MPI_CXX_FLOAT_COMPLEX       , "CXX_FLOAT_COMPLEX"       )  # noqa
cdef Datatype __CXX_DOUBLE_COMPLEX__      = def_Datatype( MPI_CXX_DOUBLE_COMPLEX      , "CXX_DOUBLE_COMPLEX"      )  # noqa
cdef Datatype __CXX_LONG_DOUBLE_COMPLEX__ = def_Datatype( MPI_CXX_LONG_DOUBLE_COMPLEX , "CXX_LONG_DOUBLE_COMPLEX" )  # noqa

cdef Datatype __SHORT_INT__        = def_Datatype( MPI_SHORT_INT       , "SHORT_INT"       )  # noqa
cdef Datatype __INT_INT__          = def_Datatype( MPI_2INT            , "INT_INT"         )  # noqa
cdef Datatype __LONG_INT__         = def_Datatype( MPI_LONG_INT        , "LONG_INT"        )  # noqa
cdef Datatype __FLOAT_INT__        = def_Datatype( MPI_FLOAT_INT       , "FLOAT_INT"       )  # noqa
cdef Datatype __DOUBLE_INT__       = def_Datatype( MPI_DOUBLE_INT      , "DOUBLE_INT"      )  # noqa
cdef Datatype __LONG_DOUBLE_INT__  = def_Datatype( MPI_LONG_DOUBLE_INT , "LONG_DOUBLE_INT" )  # noqa

cdef Datatype __CHARACTER__        = def_Datatype( MPI_CHARACTER        , "CHARACTER"        )  # noqa
cdef Datatype __LOGICAL__          = def_Datatype( MPI_LOGICAL          , "LOGICAL"          )  # noqa
cdef Datatype __INTEGER__          = def_Datatype( MPI_INTEGER          , "INTEGER"          )  # noqa
cdef Datatype __REAL__             = def_Datatype( MPI_REAL             , "REAL"             )  # noqa
cdef Datatype __DOUBLE_PRECISION__ = def_Datatype( MPI_DOUBLE_PRECISION , "DOUBLE_PRECISION" )  # noqa
cdef Datatype __COMPLEX__          = def_Datatype( MPI_COMPLEX          , "COMPLEX"          )  # noqa
cdef Datatype __DOUBLE_COMPLEX__   = def_Datatype( MPI_DOUBLE_COMPLEX   , "DOUBLE_COMPLEX"   )  # noqa

cdef Datatype __LOGICAL1__  = def_Datatype( MPI_LOGICAL1  , "LOGICAL1"  )
cdef Datatype __LOGICAL2__  = def_Datatype( MPI_LOGICAL2  , "LOGICAL2"  )
cdef Datatype __LOGICAL4__  = def_Datatype( MPI_LOGICAL4  , "LOGICAL4"  )
cdef Datatype __LOGICAL8__  = def_Datatype( MPI_LOGICAL8  , "LOGICAL8"  )
cdef Datatype __LOGICAL16__ = def_Datatype( MPI_LOGICAL16 , "LOGICAL16" )
cdef Datatype __INTEGER1__  = def_Datatype( MPI_INTEGER1  , "INTEGER1"  )
cdef Datatype __INTEGER2__  = def_Datatype( MPI_INTEGER2  , "INTEGER2"  )
cdef Datatype __INTEGER4__  = def_Datatype( MPI_INTEGER4  , "INTEGER4"  )
cdef Datatype __INTEGER8__  = def_Datatype( MPI_INTEGER8  , "INTEGER8"  )
cdef Datatype __INTEGER16__ = def_Datatype( MPI_INTEGER16 , "INTEGER16" )
cdef Datatype __REAL2__     = def_Datatype( MPI_REAL2     , "REAL2"     )
cdef Datatype __REAL4__     = def_Datatype( MPI_REAL4     , "REAL4"     )
cdef Datatype __REAL8__     = def_Datatype( MPI_REAL8     , "REAL8"     )
cdef Datatype __REAL16__    = def_Datatype( MPI_REAL16    , "REAL16"    )
cdef Datatype __COMPLEX4__  = def_Datatype( MPI_COMPLEX4  , "COMPLEX4"  )
cdef Datatype __COMPLEX8__  = def_Datatype( MPI_COMPLEX8  , "COMPLEX8"  )
cdef Datatype __COMPLEX16__ = def_Datatype( MPI_COMPLEX16 , "COMPLEX16" )
cdef Datatype __COMPLEX32__ = def_Datatype( MPI_COMPLEX32 , "COMPLEX32" )

# Predefined datatype handles
# ---------------------------

DATATYPE_NULL = __DATATYPE_NULL__  #: Null datatype handle
# MPI-specific datatypes
PACKED = __PACKED__
BYTE   = __BYTE__
AINT   = __AINT__
OFFSET = __OFFSET__
COUNT  = __COUNT__
# Elementary C datatypes
CHAR                = __CHAR__
WCHAR               = __WCHAR__
SIGNED_CHAR         = __SIGNED_CHAR__
SHORT               = __SHORT__
INT                 = __INT__
LONG                = __LONG__
LONG_LONG           = __LONG_LONG__
UNSIGNED_CHAR       = __UNSIGNED_CHAR__
UNSIGNED_SHORT      = __UNSIGNED_SHORT__
UNSIGNED            = __UNSIGNED__
UNSIGNED_LONG       = __UNSIGNED_LONG__
UNSIGNED_LONG_LONG  = __UNSIGNED_LONG_LONG__
FLOAT               = __FLOAT__
DOUBLE              = __DOUBLE__
LONG_DOUBLE         = __LONG_DOUBLE__
# C99 datatypes
C_BOOL                = __C_BOOL__
INT8_T                = __INT8_T__
INT16_T               = __INT16_T__
INT32_T               = __INT32_T__
INT64_T               = __INT64_T__
UINT8_T               = __UINT8_T__
UINT16_T              = __UINT16_T__
UINT32_T              = __UINT32_T__
UINT64_T              = __UINT64_T__
FLOAT16_T             = __FLOAT16_T__
FLOAT32_T             = __FLOAT32_T__
FLOAT64_T             = __FLOAT64_T__
BFLOAT16_T            = __BFLOAT16_T__
C_COMPLEX             = __C_COMPLEX__
C_FLOAT_COMPLEX       = __C_FLOAT_COMPLEX__
C_DOUBLE_COMPLEX      = __C_DOUBLE_COMPLEX__
C_LONG_DOUBLE_COMPLEX = __C_LONG_DOUBLE_COMPLEX__
# C++ datatypes
CXX_BOOL                = __CXX_BOOL__
CXX_FLOAT_COMPLEX       = __CXX_FLOAT_COMPLEX__
CXX_DOUBLE_COMPLEX      = __CXX_DOUBLE_COMPLEX__
CXX_LONG_DOUBLE_COMPLEX = __CXX_LONG_DOUBLE_COMPLEX__
# C Datatypes for reduction operations
SHORT_INT        = __SHORT_INT__
INT_INT = TWOINT = __INT_INT__
LONG_INT         = __LONG_INT__
FLOAT_INT        = __FLOAT_INT__
DOUBLE_INT       = __DOUBLE_INT__
LONG_DOUBLE_INT  = __LONG_DOUBLE_INT__
# Elementary Fortran datatypes
CHARACTER        = __CHARACTER__
LOGICAL          = __LOGICAL__
INTEGER          = __INTEGER__
REAL             = __REAL__
DOUBLE_PRECISION = __DOUBLE_PRECISION__
COMPLEX          = __COMPLEX__
DOUBLE_COMPLEX   = __DOUBLE_COMPLEX__
# Size-specific Fortran datatypes
LOGICAL1  = __LOGICAL1__
LOGICAL2  = __LOGICAL2__
LOGICAL4  = __LOGICAL4__
LOGICAL8  = __LOGICAL8__
LOGICAL16 = __LOGICAL16__
INTEGER1  = __INTEGER1__
INTEGER2  = __INTEGER2__
INTEGER4  = __INTEGER4__
INTEGER8  = __INTEGER8__
INTEGER16 = __INTEGER16__
REAL2     = __REAL2__
REAL4     = __REAL4__
REAL8     = __REAL8__
REAL16    = __REAL16__
COMPLEX4  = __COMPLEX4__
COMPLEX8  = __COMPLEX8__
COMPLEX16 = __COMPLEX16__
COMPLEX32 = __COMPLEX32__

# Convenience aliases
UNSIGNED_INT          = __UNSIGNED__
SIGNED_SHORT          = __SHORT__
SIGNED_INT            = __INT__
SIGNED_LONG           = __LONG__
SIGNED_LONG_LONG      = __LONG_LONG__
BOOL                  = __C_BOOL__
SINT8_T               = __INT8_T__
SINT16_T              = __INT16_T__
SINT32_T              = __INT32_T__
SINT64_T              = __INT64_T__
F_BOOL                = __LOGICAL__
F_INT                 = __INTEGER__
F_FLOAT               = __REAL__
F_DOUBLE              = __DOUBLE_PRECISION__
F_COMPLEX             = __COMPLEX__
F_FLOAT_COMPLEX       = __COMPLEX__
F_DOUBLE_COMPLEX      = __DOUBLE_COMPLEX__
